"""pytest-bdd public API."""
from pytest_bdd.packaging import get_distribution_version
from pytest_bdd.scenario import FeaturePathType, scenario, scenarios
from pytest_bdd.steps import given, step, then, when
from pytest_bdd.warning_types import PytestBDDStepDefinitionWarning

__version__ = str(get_distribution_version("pytest-bdd-ng"))

__all__ = ["given", "when", "then", "step", "scenario", "scenarios"]



import textwrap


def test_hooks(testdir):
    subdir = testdir.mkpydir("subdir")
    subdir.join("conftest.py").write(
        textwrap.dedent(
            # language=python
            r"""
            def pytest_pyfunc_call(pyfuncitem):
                print('\npytest_pyfunc_call hook')

            def pytest_generate_tests(metafunc):
                print('\npytest_generate_tests hook')
            """
        )
    )

    subdir.join("foo.feature").write(
        textwrap.dedent(
            # language=gherkin
            r"""
            Feature: The feature
                Scenario: Some scenario
            """
        )
    )

    result = testdir.runpytest("-s")

    assert result.stdout.lines.count("pytest_pyfunc_call hook") == 1
    assert result.stdout.lines.count("pytest_generate_tests hook") == 1


def test_item_collection_does_not_break_on_non_function_items(testdir):
    """Regression test for https://github.com/pytest-dev/pytest-bdd/issues/317"""
    testdir.makeconftest(
        # language=python
        """\
        import pytest

        @pytest.hookimpl(tryfirst=True)
        def pytest_collection_modifyitems(session, config, items):
            try:
                item_creator = CustomItem.from_parent  # Only available in pytest >= 5.4.0
            except AttributeError:
                item_creator = CustomItem

            items[:] = [item_creator(name=item.name, parent=item.parent) for item in items]

        class CustomItem(pytest.Item):
            def runtest(self):
                assert True
        """
    )

    testdir.makepyfile(
        # language=python
        """\
        def test_convert_me_to_custom_item_and_assert_true():
            assert False
        """
    )

    result = testdir.runpytest()
    result.assert_outcomes(passed=1)



from collections import defaultdict, namedtuple
from enum import Enum
from functools import partial
from inspect import getfile
from itertools import chain, product, starmap
from operator import attrgetter, eq, is_not
from pathlib import Path
from typing import Any, ClassVar, List, Literal, Mapping, Optional, Sequence, Type, Union

from attr import attrib, attrs
from pydantic import (  # type:ignore[attr-defined] # migration to pydantic 2
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    ValidationError,
    model_validator,
)

from messages import KeywordType, MediaType, Source  # type:ignore[attr-defined]
from pytest_bdd.compatibility.typing import Annotated, Self
from pytest_bdd.mimetypes import Mimetype
from pytest_bdd.scenario_locator import ScenarioLocatorFilterMixin
from pytest_bdd.utils import deepattrgetter

# mypy: disable-error-code="typeddict-unknown-key, typeddict-item"


class Keyword(Enum):
    Given = "Given"
    When = "When"
    Then = "Then"
    And = "And"
    But = "But"
    Star = "*"


class SubKeyword(Enum):
    Step = "Step"
    Alternative = "Alternative"


KEYWORD_TO_TYPE: Mapping[Union[Keyword, str, None], KeywordType] = defaultdict(
    lambda: KeywordType.unknown,
    [
        (Keyword.Given, KeywordType.context),
        (Keyword.When, KeywordType.action),
        (Keyword.Then, KeywordType.outcome),
        (Keyword.And, KeywordType.conjunction),
        (Keyword.But, KeywordType.conjunction),
        (Keyword.Star, KeywordType.unknown),
        (None, KeywordType.unknown),
    ],
)


class Node(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        populate_by_name=True,
    )

    tags: Optional[Sequence[str]] = Field(default_factory=list, alias="Tags")
    name: Optional[str] = Field(None, alias="Name")
    description: Optional[str] = Field(None, alias="Description")
    comments: Optional[Sequence[str]] = Field(default_factory=list, alias="Comments")


class Table(Node):
    type: Optional[Literal["Rowed", "Columned"]] = Field("Rowed", alias="Type")
    parameters: Optional[Sequence[str]] = Field(default_factory=list, alias="Parameters")
    values: Optional[Sequence[Sequence[Any]]] = Field(default_factory=list, alias="Values")

    @property
    def columned_values(self):
        return self.values if self.type == "Columned" else list(zip(*self.values))

    @property
    def rowed_values(self):
        return self.values if self.type == "Rowed" else list(zip(*self.values))


class SubTable(Node):
    sub_table: Table = Field(..., alias="Table")


@AfterValidator
def convert_sub_tables_to_tables(value):
    return value.sub_table if isinstance(value, SubTable) else value


class Join(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        populate_by_name=True,
    )

    tables: List[Annotated[Union[Table, "Join", SubTable], convert_sub_tables_to_tables]] = Field(
        default_factory=list, alias="Join"
    )

    __hash__ = id

    @property
    def tags(self):
        return list(dict.fromkeys(chain.from_iterable(map(attrgetter("tags"), self.tables))))

    @property
    def name(self):
        return "\n".join(filter(partial(is_not, None), list(chain(map(attrgetter("name"), self.tables)))))

    @property
    def description(self):
        return "\n".join(
            filter(
                partial(is_not, None),
                chain.from_iterable(map(deepattrgetter("description", skip_missing=True), self.tables)),
            )
        )

    @property
    def comments(self):
        return list(chain.from_iterable(map(attrgetter("comments"), self.tables)))

    @property
    def parameters(self):
        return list(dict.fromkeys(chain.from_iterable(map(attrgetter("parameters"), self.tables))))

    @property
    def type(self):
        return "Rowed"

    @property
    def values(self):
        def _():
            filled_tables = list(filter(attrgetter("parameters"), self.tables))
            if filled_tables:
                filled_tables_parameters = list(chain.from_iterable(map(attrgetter("parameters"), self.tables)))
                for filled_tables_values in map(
                    lambda tables_values: list(chain.from_iterable(tables_values)),
                    product(*map(attrgetter("rowed_values"), filled_tables)),
                ):
                    if all(
                        [
                            all(
                                starmap(
                                    eq,
                                    product(
                                        [
                                            value
                                            for _parameter, value in zip(filled_tables_parameters, filled_tables_values)
                                            if parameter == _parameter
                                        ],
                                        repeat=2,
                                    ),
                                )
                            )
                            for parameter in self.parameters
                        ]
                    ):

                        def values_gen():
                            for parameter in self.parameters:
                                for _parameter, value in zip(filled_tables_parameters, filled_tables_values):
                                    if parameter == _parameter:
                                        yield value
                                        break

                        values = list(values_gen())
                        yield values
            else:
                yield from map(
                    lambda values_combination: list(chain.from_iterable(values_combination)),
                    product(*map(attrgetter("rowed_values"), self.tables)),
                )

        _values = list(_())
        return _values

    @property
    def columned_values(self):
        return list(zip(*self.values))

    @property
    def rowed_values(self):
        return self.values


@BeforeValidator
def before_convert_to_step(value):
    if isinstance(value, str):
        return Step(action=value)
    elif isinstance(value, dict) and len(value) == 1 and next(iter(value)) not in SubKeyword.__members__:
        return Step(type=next(iter(value.keys())), action=next(iter(value.values())))
    else:
        return value


@AfterValidator
def select_step_keyword_type(value):
    try:
        return Keyword(value)
    except ValueError:
        return value


@AfterValidator
def after_convert_sub_steps_to_steps(value):
    return value.sub_step if isinstance(value, SubStep) else value


StepKeywordType = Union[Keyword, Annotated[str, select_step_keyword_type]]


class StepPrototype(Node):
    steps: Sequence[
        Annotated[
            Annotated[Union["SubStep", "Alternative", "StepPrototype"], before_convert_to_step],
            after_convert_sub_steps_to_steps,
        ]
    ] = Field(default_factory=list, alias="Steps")

    type: Optional[StepKeywordType] = Field(default=Keyword.Star, alias="Type")
    data: List[Annotated[Union[Table, Join, SubTable], convert_sub_tables_to_tables]] = Field(
        default_factory=list, alias="Data"
    )
    examples: List[Annotated[Union[Table, Join, SubTable], convert_sub_tables_to_tables]] = Field(
        default_factory=list, alias="Examples"
    )
    keyword_type: Optional[KeywordType] = Field(KeywordType.unknown)

    Route: ClassVar[Type] = namedtuple("Route", ["tags", "steps", "example_table"])

    @model_validator(mode="after")  # type: ignore[misc] # migration to pydantic 2
    def set_keyword_type(self) -> Self:
        self.keyword_type = KEYWORD_TO_TYPE[self.type]
        return self  # type: ignore[return-value] # migration to pydantic 2

    @property
    def routes(self):
        for routes in (
            product(*map(attrgetter("routes"), self.steps))
            if self.steps
            else [[self.Route([], [], Table(parameters=[], values=[]))]]
        ):
            steps = [self, *chain.from_iterable(map(attrgetter("steps"), routes))]

            if self.examples:
                for _example_table in self.examples:
                    example_table = Join(tables=[*map(attrgetter("example_table"), routes), _example_table])
                    tags = list(
                        {*chain.from_iterable(map(attrgetter("tags"), routes)), *example_table.tags, *self.tags}
                    )

                    yield self.Route(
                        tags,
                        steps,
                        example_table,
                    )
            else:
                example_table = Join(tables=[*map(attrgetter("example_table"), routes)])
                tags = list({*chain.from_iterable(map(attrgetter("tags"), routes)), *example_table.tags, *self.tags})

                yield self.Route(
                    tags,
                    steps,
                    example_table,
                )

    @classmethod
    def build_by_action(cls, action, *args, **kwargs):
        return cls(*args, **kwargs, action=action)

    @attrs
    class Locator(ScenarioLocatorFilterMixin):
        step: "StepPrototype" = attrib()
        filename = attrib()
        uri = attrib()
        mimetype = attrib()

        def resolve_features(self, config):
            from pytest_bdd.struct_bdd.model_builder import GherkinDocumentBuilder

            feature = GherkinDocumentBuilder(self.step).build_feature(
                filename=self.filename, uri=self.uri, id_generator=config.pytest_bdd_id_generator
            )

            if isinstance(self.mimetype, MediaType):
                media_type = self.mimetype
            elif isinstance(self.mimetype, Mimetype):
                media_type = self.mimetype.value
            else:
                media_type = str(self.mimetype)
            try:
                feature_source = Source(uri=self.uri, data=Path(self.filename).read_text(), media_type=media_type)
                yield feature, feature_source
            except ValidationError:
                # Workaround because of https://github.com/cucumber/messages/issues/161
                yield feature, None

    def as_test(self, filename):
        from pytest_bdd.scenario import scenarios

        return scenarios(
            locators=[
                self.Locator(
                    self,
                    str(Path(filename).as_posix()),
                    str(Path(filename).relative_to(Path.cwd())),
                    mimetype=Mimetype.python,
                )
            ],
            return_test_decorator=False,
        )

    def as_test_decorator(self, filename):
        from pytest_bdd.scenario import scenarios

        return scenarios(
            locators=[
                self.Locator(
                    self,
                    str(Path(filename).as_posix()),
                    str(Path(filename).relative_to(Path.cwd())),
                    mimetype=Mimetype.python,
                )
            ],
            return_test_decorator=True,
        )

    def __call__(self, func):
        return self.as_test_decorator(getfile(func))(func)


class Alternative(Node):
    steps: Sequence[
        Annotated[
            Annotated[Union["SubStep", "Alternative", "StepPrototype"], before_convert_to_step],
            after_convert_sub_steps_to_steps,
        ]
    ] = Field(default_factory=list, alias="Alternative")

    @property
    def routes(self):
        yield from chain.from_iterable(map(attrgetter("routes"), self.steps))


class Step(StepPrototype):
    type: Optional[StepKeywordType] = Field(default=Keyword.Star, alias="Type")
    action: Optional[str] = Field(None, alias="Action")


class SubStep(BaseModel):
    sub_step: Step = Field(..., alias="Step")


class StarStep(StepPrototype):
    type: StepKeywordType = Field(Keyword.Star, alias="Type")
    action: Optional[str] = Field(alias=Keyword.Star.value)


class GivenStep(StepPrototype):
    type: StepKeywordType = Field(Keyword.Given, alias="Type")
    action: Optional[str] = Field(alias=Keyword.Given.value)


class WhenStep(StepPrototype):
    type: StepKeywordType = Field(Keyword.When, alias="Type")
    action: Optional[str] = Field(alias=Keyword.When.value)


class ThenStep(StepPrototype):
    type: StepKeywordType = Field(Keyword.Then, alias="Type")
    action: Optional[str] = Field(alias=Keyword.Then.value)


class AndStep(StepPrototype):
    type: StepKeywordType = Field(Keyword.And, alias="Type")
    action: Optional[str] = Field(alias=Keyword.And.value)


class ButStep(StepPrototype):
    type: StepKeywordType = Field(Keyword.But, alias="Type")
    action: Optional[str] = Field(alias=Keyword.But.value)


Join.model_rebuild()  # type:ignore[attr-defined] # migration to pydantic 2
StepPrototype.model_rebuild()  # type:ignore[attr-defined] # migration to pydantic 2
Alternative.model_rebuild()  # type:ignore[attr-defined] # migration to pydantic 2

Given = GivenStep.build_by_action
When = WhenStep.build_by_action
Then = ThenStep.build_by_action
And = AndStep.build_by_action
But = ButStep.build_by_action



import os
import sys


def relpath(path, start=os.curdir):
    try:
        return os.path.relpath(path, start)
    except ValueError:
        if sys.platform == "win32":
            return path
        else:
            raise



from contextlib import contextmanager
from enum import Enum
from inspect import isfunction, isgeneratorfunction, signature
from itertools import count, product, starmap
from typing import Optional, Union

from _pytest.mark import Mark
from decopatch import function_decorator
from makefun import wraps
from pytest import fixture

from pytest_bdd.compatibility.pytest import PYTEST7, FixtureRequest
from pytest_bdd.tag_expression import GherkinTagExpression, MarksTagExpression, TagExpression

expression_count_gen = count()


class HookKind(Enum):
    mark = "mark"
    tag = "tag"


class HookConjunction(Enum):
    before = "before"
    after = "after"
    around = "around"


def decorator_builder(conjunction: Union[str, HookConjunction], kind: Union[str, HookKind]):
    _conjunction = HookConjunction(conjunction) if isinstance(conjunction, str) else conjunction
    _kind = HookKind(kind) if isinstance(kind, str) else kind

    @function_decorator
    def decorator_wrapper(expression: Optional[str] = None, name: Optional[str] = None):
        _expression: str = expression if expression is not None else ""

        def decorator(func):
            func_sig = signature(func)

            fixture_decorator = fixture(
                name=f"{_conjunction.value}_{_kind.value}_expression_{_expression}_{next(expression_count_gen)}",
                autouse=True,
            )

            @wraps(func, prepend_args="request", remove_args="request")
            def hook(request: FixtureRequest, *args, **kwargs):
                _ExpressionType = {
                    HookKind.mark: MarksTagExpression,
                    HookKind.tag: GherkinTagExpression,
                }[_kind]

                # mypy@Python 3.8 complains "ABCMeta" has no attribute "parse"  [attr-defined] what is wrong
                parsed_expression: TagExpression = _ExpressionType.parse(_expression)  # type: ignore[attr-defined]

                get_marks = lambda: list(
                    {
                        HookKind.mark: request.node.iter_markers(),
                        HookKind.tag: map(
                            lambda tag: Mark(  # type: ignore[no-any-return]
                                tag.name, args=tuple(), kwargs={}, **({"_ispytest": True} if PYTEST7 else {})
                            ),
                            request.getfixturevalue("scenario").tags,
                        ),
                    }[_kind]
                )

                is_matching = parsed_expression.evaluate(get_marks())

                is_function = isfunction(func)
                is_generator_function = isgeneratorfunction(func)
                if any(
                    [
                        _conjunction is HookConjunction.around and not is_generator_function,
                        _conjunction in {HookConjunction.before, HookConjunction.after} and not is_function,
                    ]
                ):
                    raise ValueError(f"_{_conjunction.value}")

                _args, _kwargs = args, {
                    **kwargs,
                    **({"request": request} if "request" in func_sig.parameters.keys() else {}),
                }

                if is_matching:
                    if _conjunction is HookConjunction.before:
                        yield func(*_args, **_kwargs)
                    elif _conjunction is HookConjunction.after:
                        yield
                        func(*_args, **_kwargs)
                    elif _conjunction is HookConjunction.around:
                        with contextmanager(func)(*_args, **_kwargs):
                            yield
                    else:  # pragma: no cover
                        yield
                else:
                    yield

            hook.__pytest_bdd_is_hook__ = True
            if name is not None:
                hook.__pytest_bdd_hook_name__ = name
            hook.__pytest_bdd_hook_expression__ = _expression

            return fixture_decorator(hook)

        return decorator

    return decorator_wrapper


before_mark, before_tag, after_mark, after_tag, around_mark, around_tag = starmap(
    decorator_builder, product(HookConjunction, HookKind)
)



import os
import plistlib


def read(path):
    """return webloc url"""
    if hasattr(plistlib, "load"):
        with open(path, "rb") as f:
            return plistlib.load(f).get("URL")
    return plistlib.readPlist(path).get("URL")


def write(path, url):
    """write url to webloc file"""
    data = dict(URL=str(url))
    dir = os.path.dirname(path)
    if not os.path.exists(dir):
        os.makedirs(dir)
    if hasattr(plistlib, "dump"):
        with open(path, "wb") as f:
            plistlib.dump(data, f)
    else:
        plistlib.writePlist(data, path)



"""StepHandler decorators.

Example:

@given("I have an article", target_fixture="article")
def given_article(author):
    return create_test_article(author=author)


@when("I go to the article page")
def go_to_the_article_page(browser, article):
    browser.visit(urljoin(browser.url, "/articles/{0}/".format(article.id)))


@then("I should not see the error message")
def no_error_message(browser):
    with pytest.raises(ElementDoesNotExist):
        browser.find_by_css(".message.error").first


Multiple names for the steps:

@given("I have an article")
@given("there is an article")
def article(author):
    return create_test_article(author=author)


Reusing existing fixtures for a different step name:


@given("I have a beautiful article")
def given_beautiful_article(article):
    pass

"""
import warnings
from contextlib import suppress
from inspect import getfile, getsourcelines
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    Set,
    Type,
    Union,
    cast,
)
from uuid import uuid4
from warnings import warn

import pytest
from _pytest.fixtures import FixtureRequest
from attr import Factory, attrib, attrs
from ordered_set import OrderedSet
from pydantic import ValidationError

from messages import ExpressionType, Location, Pickle  # type:ignore[attr-defined]
from messages import PickleStep as Step  # type:ignore[attr-defined]
from messages import SourceReference, StepDefinition, StepDefinitionPattern  # type:ignore[attr-defined]
from pytest_bdd.compatibility.path import relpath
from pytest_bdd.compatibility.pytest import Config, FixtureLookupError, Parser, TypeAlias, get_config_root_path
from pytest_bdd.model import Feature, StepType
from pytest_bdd.model.messages_extension import ExpressionType as ExpressionTypeExtension
from pytest_bdd.parsers import StepParser
from pytest_bdd.utils import (
    PytestBDDIdGeneratorHandler,
    convert_str_to_python_name,
    get_caller_module_locals,
    getitemdefault,
    setdefaultattr,
)
from pytest_bdd.warning_types import PytestBDDStepDefinitionWarning


def add_options(parser: Parser):
    """Add pytest-bdd options."""
    group = parser.getgroup("bdd", "Steps")
    group.addoption(
        "--liberal-steps",
        action="store_true",
        dest="liberal_steps",
        default=None,
        help="Allow use different keywords with same step definition",
    )
    parser.addini(
        "liberal_steps",
        default=False,
        type="bool",
        help="Allow use different keywords with same step definition",
    )


def given(
    parserlike: Any,
    anonymous_group_names: Optional[Iterable[str]] = None,
    converters: Optional[Dict[str, Callable]] = None,
    target_fixture: Optional[str] = None,
    target_fixtures: Optional[Sequence[str]] = None,
    params_fixtures_mapping: Union[Set[str], Dict[str, str], Any] = True,
    param_defaults: Optional[dict] = None,
    liberal: Optional[bool] = None,
    stacklevel=1,
) -> Callable:
    """Given step decorator.

    :param parserlike: StepHandler name or a parser object.
    :param anonymous_group_names: Grant names for anonymous groups of parserlike
    :param converters: Optional `dict` of the argument or parameter converters in form
                       {<param_name>: <converter function>}.
    :param target_fixture: Target fixture name to replace by steps definition function.
    :param target_fixtures: Target fixture names to be replaced by steps definition function.
    :param params_fixtures_mapping: StepHandler parameters would be injected as fixtures
    :param param_defaults: Default parameters for step definition
    :param liberal: Could step definition be used with other keywords
    :param stacklevel: Stack level to find the caller frame. This is used when injecting the step definition fixture.


    :return: Decorator function for the step.
    """
    return StepHandler.decorator_builder(
        StepType.context,
        parserlike,
        anonymous_group_names=anonymous_group_names,
        converters=converters,
        target_fixture=target_fixture,
        target_fixtures=target_fixtures,
        params_fixtures_mapping=params_fixtures_mapping,
        param_defaults=param_defaults,
        liberal=liberal,
        stacklevel=stacklevel + 1,
    )


def when(
    parserlike: Any,
    anonymous_group_names: Optional[Iterable[str]] = None,
    converters: Optional[Dict[str, Callable]] = None,
    target_fixture: Optional[str] = None,
    target_fixtures: Optional[Sequence[str]] = None,
    params_fixtures_mapping: Union[Set[str], Dict[str, str], Any] = True,
    param_defaults: Optional[dict] = None,
    liberal: Optional[bool] = None,
    stacklevel=1,
) -> Callable:
    """When step decorator.

    :param parserlike: StepHandler name or a parser object.
    :param anonymous_group_names: Grant names for anonymous groups of parserlike
    :param converters: Optional `dict` of the argument or parameter converters in form
                       {<param_name>: <converter function>}.
    :param target_fixture: Target fixture name to replace by steps definition function.
    :param target_fixtures: Target fixture names to be replaced by steps definition function.
    :param params_fixtures_mapping: StepHandler parameters would be injected as fixtures
    :param param_defaults: Default parameters for step definition
    :param liberal: Could step definition be used with other keywords
    :param stacklevel: Stack level to find the caller frame. This is used when injecting the step definition fixture.

    :return: Decorator function for the step.
    """
    return StepHandler.decorator_builder(
        StepType.action,
        parserlike,
        anonymous_group_names=anonymous_group_names,
        converters=converters,
        target_fixture=target_fixture,
        target_fixtures=target_fixtures,
        params_fixtures_mapping=params_fixtures_mapping,
        param_defaults=param_defaults,
        liberal=liberal,
        stacklevel=stacklevel + 1,
    )


def then(
    parserlike: Any,
    anonymous_group_names: Optional[Iterable[str]] = None,
    converters: Optional[Dict[str, Callable]] = None,
    target_fixture: Optional[str] = None,
    target_fixtures: Optional[Sequence[str]] = None,
    params_fixtures_mapping: Union[Set[str], Dict[str, str], Any] = True,
    param_defaults: Optional[dict] = None,
    liberal: Optional[bool] = None,
    stacklevel=1,
) -> Callable:
    """Then step decorator.

    :param parserlike: StepHandler name or a parser object.
    :param anonymous_group_names: Grant names for anonymous groups of parserlike
    :param converters: Optional `dict` of the argument or parameter converters in form
                       {<param_name>: <converter function>}.
    :param target_fixture: Target fixture name to replace by steps definition function.
    :param target_fixtures: Target fixture names to be replaced by steps definition function.
    :param params_fixtures_mapping: StepHandler parameters would be injected as fixtures
    :param param_defaults: Default parameters for step definition
    :param liberal: Could step definition be used with other keywords
    :param stacklevel: Stack level to find the caller frame. This is used when injecting the step definition fixture.

    :return: Decorator function for the step.
    """
    return StepHandler.decorator_builder(
        StepType.outcome,
        parserlike,
        anonymous_group_names=anonymous_group_names,
        converters=converters,
        target_fixture=target_fixture,
        target_fixtures=target_fixtures,
        params_fixtures_mapping=params_fixtures_mapping,
        param_defaults=param_defaults,
        liberal=liberal,
        stacklevel=stacklevel + 1,
    )


def step(
    parserlike: Any,
    anonymous_group_names: Optional[Iterable[str]] = None,
    converters: Optional[Dict[str, Callable]] = None,
    target_fixture: Optional[str] = None,
    target_fixtures: Optional[Sequence[str]] = None,
    params_fixtures_mapping: Union[Set[str], Dict[str, str], Any] = True,
    param_defaults: Optional[dict] = None,
    liberal: Optional[bool] = None,
    stacklevel=1,
):
    """Liberal step decorator which could be used with any keyword.

    :param parserlike: StepHandler name or a parser object.
    :param anonymous_group_names: Grant names for anonymous groups of parserlike
    :param converters: Optional `dict` of the argument or parameter converters in form
                       {<param_name>: <converter function>}.
    :param target_fixture: Target fixture name to replace by steps definition function.
    :param target_fixtures: Target fixture names to be replaced by steps definition function.
    :param params_fixtures_mapping: StepHandler parameters would be injected as fixtures
    :param param_defaults: Default parameters for step definition
    :param liberal: Could step definition be used with other keywords
    :param stacklevel: Stack level to find the caller frame. This is used when injecting the step definition fixture.

    :return: Decorator function for the step.
    """
    return StepHandler.decorator_builder(
        StepType.unknown,
        parserlike,
        anonymous_group_names=anonymous_group_names,
        converters=converters,
        target_fixture=target_fixture,
        target_fixtures=target_fixtures,
        params_fixtures_mapping=params_fixtures_mapping,
        param_defaults=param_defaults,
        liberal=liberal,
        stacklevel=stacklevel + 1,
    )


class StepHandler:
    Model: TypeAlias = "Step"

    @attrs
    class Matcher:
        config: Config = attrib()
        feature: Feature = attrib(init=False)
        pickle: Pickle = attrib(init=False)
        step: Step = attrib(init=False)
        previous_step: Optional[Step] = attrib(init=False)
        step_registry: "StepHandler.Registry" = attrib(init=False)
        step_type_context = attrib(default=None)

        class MatchNotFoundError(RuntimeError):
            pass

        def __call__(
            self,
            request,
            feature: Feature,
            pickle: Pickle,
            step: Step,
            previous_step: Optional[Step],
            step_registry: "StepHandler.Registry",
        ) -> "StepHandler.Definition":
            self.request = request
            self.feature = feature
            self.pickle = pickle
            self.step = step
            self.previous_step = previous_step
            self.step_registry = step_registry

            self.step_type_context = (
                self.step_type_context
                if self.step.type is StepType.unknown and self.step_type_context is not None
                else self.step.type
            )

            step_definitions = list(
                self.find_step_definition_matches(
                    self.step_registry, (self.strict_matcher, self.unspecified_matcher, self.liberal_matcher)
                )
            )

            if len(step_definitions) > 0:
                if len(step_definitions) > 1:
                    warn(PytestBDDStepDefinitionWarning(f"Alternative step definitions are found: {step_definitions}"))
                return step_definitions[0]
            raise self.MatchNotFoundError(self.step.text)

        def strict_matcher(self, step_definition):
            return step_definition.type_ == self.step_type_context and step_definition.parser.is_matching(
                self.request,
                self.step.text,
            )

        def unspecified_matcher(self, step_definition):
            return (
                self.step_type_context == StepType.unknown or step_definition.type_ == StepType.unknown
            ) and step_definition.parser.is_matching(
                self.request,
                self.step.text,
            )

        def liberal_matcher(self, step_definition):
            if step_definition.liberal is None:
                if self.config.option.liberal_steps is not None:
                    is_step_definition_liberal = self.config.option.liberal_steps
                else:
                    is_step_definition_liberal = self.config.getini("liberal_steps")
            else:
                is_step_definition_liberal = step_definition.liberal

            return all(
                (
                    not self.unspecified_matcher(step_definition),
                    is_step_definition_liberal,
                    step_definition.type_ != self.step_type_context,
                    step_definition.parser.is_matching(self.request, self.step.text),
                )
            )

        @staticmethod
        def find_step_definition_matches(
            registry: Optional["StepHandler.Registry"], matchers: Sequence[Callable[["StepHandler.Definition"], bool]]
        ) -> Iterable["StepHandler.Definition"]:
            if registry:
                found_matches = False
                for matcher in matchers:
                    for step_definition in registry:
                        if matcher(step_definition):
                            found_matches = True
                            yield step_definition
                    if found_matches:
                        break
                if not found_matches:
                    with suppress(AttributeError):
                        yield from StepHandler.Matcher.find_step_definition_matches(registry.parent, matchers)

    @attrs(eq=False)
    class Definition:
        func: Callable = attrib()
        type_: Optional[Union[str, StepType]] = attrib()
        parser: StepParser = attrib()
        anonymous_group_names: Optional[Collection[str]] = attrib()
        converters: Dict[str, Callable] = attrib()
        params_fixtures_mapping: Union[  # type: ignore[valid-type]
            Collection[str], Mapping[Union[str, Any], Union[str, Any, None]], Any
        ] = attrib()
        param_defaults: dict = attrib()
        target_fixtures: Sequence[str] = attrib()
        liberal: Optional[Any] = attrib()

        id = attrib(init=False)
        __cache = attrib(default=Factory(dict))

        @property
        def fixtures_mapped_from_step_definition(self):
            known_params = {
                *([] if self.anonymous_group_names is None else self.anonymous_group_names),
                *([] if self.parser.arguments is None else self.parser.arguments),
            }

            fixture_names = {*self.target_fixtures}

            if isinstance(self.params_fixtures_mapping, Mapping):
                converted_params = {*filter(lambda param: param is not ..., self.params_fixtures_mapping.keys())}
                fixture_names.update(
                    filter(
                        lambda fixture_name: fixture_name is not None and fixture_name is not ...,
                        self.params_fixtures_mapping.values(),
                    )
                )
                wildcard_params_strategy = getitemdefault(self.params_fixtures_mapping, ..., default=...)
            elif isinstance(self.params_fixtures_mapping, Collection):
                converted_params = set()
                wildcard_params_strategy = None
                fixture_names.update(self.params_fixtures_mapping)
            elif bool(self.params_fixtures_mapping):
                converted_params = set()
                wildcard_params_strategy = ...
            else:
                converted_params = set()
                wildcard_params_strategy = None

            if wildcard_params_strategy is ...:
                bypassed_params = known_params.difference(converted_params)
                fixture_names.update(bypassed_params)
            return fixture_names

        def as_message(self, config: Union[Config, PytestBDDIdGeneratorHandler]):
            id_generator = cast(PytestBDDIdGeneratorHandler, config).pytest_bdd_id_generator
            try:
                message = self.__cache[id(id_generator)]
            except KeyError:
                self.id = id_generator.get_next_id()

                parser_expression_type = self.parser.type

                expression_type: Union[ExpressionType, str]
                if isinstance(parser_expression_type, ExpressionType):
                    expression_type = parser_expression_type
                elif isinstance(parser_expression_type, ExpressionTypeExtension):
                    expression_type = parser_expression_type.value
                else:
                    expression_type = str(parser_expression_type)

                try:
                    pattern = StepDefinitionPattern(source=str(self.parser), type=expression_type)
                except ValidationError:
                    # Workaround because of https://github.com/cucumber/messages/issues/160
                    pattern = StepDefinitionPattern(source=str(self.parser), type=ExpressionType.regular_expression)
                message = self.__cache[id(id_generator)] = StepDefinition(
                    id=self.id,
                    pattern=pattern,
                    source_reference=SourceReference(  # type: ignore[call-arg] # migration to pydantic2
                        uri=relpath(
                            getfile(self.func),
                            str(get_config_root_path(cast(Config, config))),
                        ),
                        location=Location(line=getsourcelines(self.func)[1]),
                    ),
                )
            return message

        def get_parameters(self, request: FixtureRequest, step: Step):
            parsed_arguments = (
                self.parser.parse_arguments(request, step.text, anonymous_group_names=self.anonymous_group_names) or {}
            )
            return {
                **self.param_defaults,
                **{arg: self.converters.get(arg, lambda _: _)(value) for arg, value in parsed_arguments.items()},
            }

    @attrs
    class Registry:
        registry: Set["StepHandler.Definition"] = attrib(default=Factory(set))
        parent: "StepHandler.Registry" = attrib(default=None, init=False)

        @classmethod
        def inject_registry_fixture_and_register_steps(cls, obj):
            steps = [
                step_candidate
                for step_candidate in obj.__dict__.values()
                if hasattr(step_candidate, "__pytest_bdd_step_definitions__")
            ]
            if not steps:
                return
            setdefaultattr(obj, "step_registry", value_factory=lambda: StepHandler.Registry().fixture)
            obj.step_registry.__pytest_bdd_step_registry__.register_steps(steps)

            for step_func in steps:
                for step_definition in step_func.__pytest_bdd_step_definitions__:
                    step_definition: StepHandler.Definition = step_definition
                    for fixture_name in step_definition.fixtures_mapped_from_step_definition:

                        @pytest.fixture
                        def _(request):
                            try:
                                return request.getfixturevalue(fixture_name)
                            except FixtureLookupError:
                                ...

                        setdefaultattr(obj, fixture_name, value_factory=lambda: (_))

        def register_step_definition(self, step_definition):
            self.registry.add(step_definition)

        def register_steps(self, step_funcs):
            for step_func in step_funcs:
                for step_definition in step_func.__pytest_bdd_step_definitions__:
                    self.register_step_definition(step_definition)

        @property
        def fixture(self):
            @pytest.fixture
            def step_registry(step_registry):
                self.parent = step_registry
                return self

            step_registry.__pytest_bdd_step_registry__ = self
            return step_registry

        def __iter__(self) -> Iterator["StepHandler.Definition"]:
            return iter(self.registry)

    @staticmethod
    def decorator_builder(
        step_type: Optional[Union[str, StepType]],
        step_parserlike: Any,
        anonymous_group_names: Optional[Iterable[str]] = None,
        converters: Optional[Dict[str, Callable]] = None,
        target_fixture: Optional[str] = None,
        target_fixtures: Optional[Sequence[str]] = None,
        params_fixtures_mapping: Union[Set[str], Dict[str, str], Any] = True,
        param_defaults: Optional[dict] = None,
        liberal: Optional[Any] = None,
        stacklevel=2,
    ) -> Callable:
        """StepHandler decorator for the type and the name.

        :param step_type: StepHandler type (CONTEXT, ACTION or OUTCOME).
        :param step_parserlike: StepHandler name as in the feature file.
        :param anonymous_group_names: Grant names for anonymous groups of parserlike
        :param converters: Optional step arguments converters mapping
        :param target_fixture: Optional fixture name to replace by step definition
        :param target_fixtures: Target fixture names to be replaced by steps definition function.
        :param params_fixtures_mapping: StepHandler parameters would be injected as fixtures
        :param param_defaults: Default parameters for step definition
        :param liberal: Could step definition be used with other keywords
        :param stacklevel: Stack level to find the caller frame. This is used when injecting the step definition fixture

        :return: Decorator function for the step.
        """

        converters = converters or {}
        param_defaults = param_defaults or {}
        if target_fixture is not None and target_fixtures is not None:
            warnings.warn(PytestBDDStepDefinitionWarning("Both target_fixture and target_fixtures are specified"))
        target_fixtures = list(
            OrderedSet(
                [
                    *([target_fixture] if target_fixture is not None else []),
                    *(target_fixtures if target_fixtures is not None else []),
                ]
            )
        )

        def decorator(step_func: Callable) -> Callable:
            """
            StepHandler decorator

            :param function step_func: StepHandler definition function
            """

            step_definition = StepHandler.Definition(  # type: ignore[call-arg]
                func=step_func,
                type_=step_type,
                parser=StepParser.build(step_parserlike),
                anonymous_group_names=anonymous_group_names,
                converters=cast(dict, converters),
                params_fixtures_mapping=params_fixtures_mapping,
                param_defaults=cast(dict, param_defaults),
                target_fixtures=cast(list, target_fixtures),
                liberal=liberal,
            )

            setdefaultattr(step_func, "__pytest_bdd_step_definitions__", value_factory=set).add(step_definition)

            # Allow step function have same names, so injecting same steps with generated names into module scope
            converted_name = convert_str_to_python_name(f'step_{step_type or ""}_{step_parserlike}_{uuid4()}')
            get_caller_module_locals(stacklevel=stacklevel)[converted_name] = step_func

            return step_func

        return decorator



import asyncio
import os
import ssl
from contextlib import suppress
from functools import partial, reduce
from itertools import filterfalse
from operator import methodcaller, truediv
from os.path import commonpath
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Callable, Iterable, Optional, Protocol, Tuple, Type, Union, cast, runtime_checkable
from urllib.parse import urljoin

import aiohttp
import certifi
from _pytest.config import Config
from attr import Factory, attrib, attrs
from pydantic import ValidationError

from messages import Source  # type:ignore[attr-defined]
from pytest_bdd.compatibility.parser import ParserProtocol
from pytest_bdd.compatibility.pytest import get_config_root_path
from pytest_bdd.mimetypes import Mimetype
from pytest_bdd.model import Feature, Pickle
from pytest_bdd.scenario import Args
from pytest_bdd.utils import PytestBDDIdGeneratorHandler, is_local_url


@runtime_checkable
class ScenarioLocatorFeatureResolver(Protocol):
    def resolve_features(
        self, config: Union[Config, PytestBDDIdGeneratorHandler]
    ) -> Iterable[Tuple[Feature, Source]]:  # pragma: no cover
        ...


@runtime_checkable
class ScenarioLocatorResolver(Protocol):
    def resolve(
        self, config: Union[Config, PytestBDDIdGeneratorHandler]
    ) -> Iterable[Tuple[Feature, Pickle, Source]]:  # pragma: no cover
        ...


@attrs
class ScenarioLocatorFilterMixin(ScenarioLocatorFeatureResolver, ScenarioLocatorResolver):
    filter_: Optional[Callable[[Config, Feature, Pickle], Tuple[Feature, Pickle]]] = attrib(default=None, kw_only=True)

    def filter_scenarios(self, feature, config):
        return (
            (feature, pickle)
            for pickle in feature.pickles
            if self.filter_ is None or self.filter_(config, feature, pickle)
        )  # type: ignore

    def resolve(self, config: Config):
        for feature, feature_data in self.resolve_features(config):
            for _, pickle in self.filter_scenarios(feature, config):
                yield feature, pickle, feature_data


@attrs
class UrlScenarioLocator(ScenarioLocatorFilterMixin):
    url_paths = attrib()
    encoding = attrib()
    features_base_url = attrib()
    mimetype = attrib()
    parser_type = attrib()
    parse_args = attrib()

    async def fetch(self, session: aiohttp.ClientSession, url):
        sslcontext = ssl.create_default_context(cafile=certifi.where())
        async with session.get(url, ssl=sslcontext) as response:
            return response.content_type, await response.text(encoding=self.encoding)

    async def fetch_all(self, urls):
        async with aiohttp.ClientSession() as session:
            return await asyncio.gather(*[self.fetch(session, url) for url in urls], return_exceptions=True)

    def resolve_features(self, config: Union[Config, PytestBDDIdGeneratorHandler]):
        urls = [*filterfalse(is_local_url, self.url_paths)]
        if self.features_base_url is not None:
            urls.extend(map(partial(urljoin, f"{self.features_base_url}/"), filter(is_local_url, self.url_paths)))
        if not urls:
            return
        loop = asyncio.new_event_loop()
        responses = loop.run_until_complete(self.fetch_all(urls))

        # Wait 250 ms for the underlying SSL connections to close
        loop.run_until_complete(asyncio.sleep(0.250))
        loop.close()

        hook_handler = cast(Config, config).hook
        encoding = self.encoding

        for url, response in zip(urls, responses):
            if isinstance(response, Exception):
                continue

            mimetype, feature_content = response

            if self.mimetype is not None:
                mimetype = self.mimetype

                if isinstance(mimetype, Mimetype):
                    mimetype = mimetype.value

            if self.parser_type is None:
                parser_type = hook_handler.pytest_bdd_get_parser(
                    config=config,
                    mimetype=mimetype,
                )
            else:
                parser_type = self.parser_type

            if parser_type is None:
                break

            parser = parser_type(id_generator=cast(PytestBDDIdGeneratorHandler, config).pytest_bdd_id_generator)

            try:
                filename = None
                with NamedTemporaryFile(mode="w", delete=False) as f:
                    filename = f.name
                    f.write(feature_content)

                feature, feature_data = parser.parse(
                    config,
                    Path(filename),
                    url,
                    *self.parse_args.args,
                    **{**dict(encoding=encoding), **self.parse_args.kwargs},
                )
                try:
                    yield feature, Source(uri=url, data=feature_data, media_type=mimetype)  # type: ignore[call-arg] # migration to pydantic2
                except ValidationError as e:
                    # Workaround because of https://github.com/cucumber/messages/issues/161
                    yield feature, None
            finally:
                if filename is not None:
                    with suppress(Exception):
                        os.unlink(filename)


@attrs
class FileScenarioLocator(ScenarioLocatorFilterMixin):
    feature_paths = attrib(default=Factory(list))
    encoding = attrib(default="utf-8")
    features_base_dir: Optional[Union[str, Path]] = attrib(default=None)
    mimetype: Optional[str] = attrib(default=None)
    parser_type: Optional[Type[ParserProtocol]] = attrib(default=None)
    parse_args: Args = attrib(default=Factory(lambda: Args((), {})))

    def _resolve_features_base_dir(self, config: Union[Config, PytestBDDIdGeneratorHandler]):
        try:
            if self.features_base_dir is None:
                features_base_dir = cast(Config, config).getini("bdd_features_base_dir")
            else:
                features_base_dir = self.features_base_dir
        except (ValueError, KeyError):
            features_base_dir = get_config_root_path(cast(Config, config))
        else:
            if callable(features_base_dir):
                features_base_dir = features_base_dir(config)

            features_base_dir = (get_config_root_path(cast(Config, config)) / Path(features_base_dir)).resolve()

        return features_base_dir

    def _gen_feature_paths(self, features_base_dir):
        for feature_pathlike in self.feature_paths:
            if isinstance(feature_pathlike, Path):
                feature_path = features_base_dir / feature_pathlike
                if feature_path.is_dir():
                    yield from filter(methodcaller("is_file"), feature_path.glob("**/*"))
                else:
                    yield feature_path
            else:
                try:
                    yield from filter(methodcaller("is_file"), features_base_dir.glob(str(feature_pathlike)))
                except IndexError:
                    yield from filter(methodcaller("is_file"), features_base_dir.glob("**/*"))

    @staticmethod
    def _build_file_uri(features_base_dir: Path, feature_path: Path):
        if feature_path.is_absolute():
            try:
                common_path = Path(commonpath([feature_path, features_base_dir]))
            except ValueError:
                rel_feature_path = feature_path
            else:
                sub_levels = len(features_base_dir.relative_to(common_path).parts)
                sub_path = reduce(truediv, [".."] * sub_levels, Path())
                rel_feature_path = sub_path / feature_path.relative_to(common_path)
        else:
            rel_feature_path = feature_path

        return "file:" + str(rel_feature_path.as_posix())

    def resolve_features(self, config: Union[Config, PytestBDDIdGeneratorHandler]):
        features_base_dir = self._resolve_features_base_dir(config)
        already_resolved_feature_paths = set()

        for feature_path in self._gen_feature_paths(features_base_dir=features_base_dir):
            feature_path_key = str(feature_path)
            if feature_path_key in already_resolved_feature_paths:
                break

            uri = self._build_file_uri(features_base_dir, feature_path)
            already_resolved_feature_paths.add(feature_path_key)
            hook_handler = cast(Config, config).hook
            encoding = self.encoding

            if self.mimetype is None:
                media_type = hook_handler.pytest_bdd_get_mimetype(config=config, path=feature_path)
            else:
                media_type = self.mimetype

            if self.parser_type is None:
                parser_type = hook_handler.pytest_bdd_get_parser(
                    config=config,
                    mimetype=media_type,
                )
            else:
                parser_type = self.parser_type

            if parser_type is None:
                break

            parser = parser_type(id_generator=cast(PytestBDDIdGeneratorHandler, config).pytest_bdd_id_generator)

            feature, feature_data = parser.parse(
                config,
                feature_path,
                uri,
                *self.parse_args.args,
                **{**dict(encoding=encoding), **self.parse_args.kwargs},
            )
            try:
                yield feature, Source(uri=uri, data=feature_data, media_type=media_type)  # type: ignore[call-arg] # migration to pydantic2
            except ValidationError as e:
                # Workaround because of https://github.com/cucumber/messages/issues/161
                yield feature, None



from enum import Enum


class ExpressionType(Enum):
    pytest_bdd_heuristic_expression = "PYTEST_BDD_HEURISTIC_EXPRESSION"
    pytest_bdd_string_expression = "PYTEST_BDD_STRING_EXPRESSION"
    pytest_bdd_regular_expression = "PYTEST_BDD_REGULAR_EXPRESSION"
    pytest_bdd_parse_expression = "PYTEST_BDD_PARSE_EXPRESSION"
    pytest_bdd_cfparse_expression = "PYTEST_BDD_CFPARSE_EXPRESSION"
    pytest_bdd_other_expression = "PYTEST_BDD_OTHER_EXPRESSION"



from messages import Pickle, PickleStep, Step, Tag  # type:ignore[attr-defined]
from messages import Type as StepType  # type:ignore[attr-defined]
from pytest_bdd.model.gherkin_document import Feature

__all__ = ["Feature", "Pickle", "PickleStep", "Step", "StepType", "Tag"]



import sys

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

__all__ = [
    "Annotated",
    "Self",
    "TypeAlias",
]



from functools import partial
from itertools import islice
from operator import attrgetter
from typing import TYPE_CHECKING, Optional

from pytest_bdd.utils import compose

if TYPE_CHECKING:  # pragma: no cover
    from cucumber_messages import DataTable


def data_table_to_dicts(data_table: Optional["DataTable"]):
    if data_table is None:
        return {}

    data_table_keys = map(attrgetter("value"), map(compose(next, iter), map(attrgetter("cells"), data_table.rows)))
    data_table_values = map(
        compose(list, compose(partial(map, attrgetter("value")), lambda items: islice(items, 1, None), iter)),
        map(attrgetter("cells"), data_table.rows),
    )
    return dict(zip(data_table_keys, data_table_values))



from pathlib import Path
from typing import TYPE_CHECKING, Callable, Optional, Protocol, Sequence, Tuple, Union, runtime_checkable

from attr import attrib, attrs

from pytest_bdd.compatibility.pytest import Config
from pytest_bdd.utils import IdGenerator, PytestBDDIdGeneratorHandler

if TYPE_CHECKING:  # pragma: no cover
    from pytest_bdd.model import Feature


@runtime_checkable
@attrs
class ParserProtocol(Protocol):
    id_generator: Optional[IdGenerator] = attrib(default=None, kw_only=True)
    # Defines which files would be parsed
    glob: Callable[[Path], Sequence[Union[str, Path]]]

    def parse(
        self, config: Union[Config, PytestBDDIdGeneratorHandler], path: Path, uri: str, *args, **kwargs
    ) -> Tuple["Feature", str]:  # pragma: no cover
        ...



import json
import logging
import os
import sys
from base64 import b64encode
from inspect import getfile, getsourcelines
from io import BufferedIOBase, TextIOBase
from pathlib import Path
from platform import machine, processor, system, version
from pprint import pformat
from queue import Empty, Queue
from threading import Event, Thread
from time import sleep, time_ns
from typing import Callable, Set, Union, cast

from attr import attrib, attrs
from ci_environment import detect_ci_environment
from cucumber_expressions.parameter_type_registry import ParameterTypeRegistry
from filelock import FileLock
from pydantic import ValidationError
from pytest import ExitCode, Session, hookimpl

from messages import Attachment, Ci, ContentEncoding, Duration  # type:ignore[attr-defined]
from messages import Envelope as Message  # type:ignore[attr-defined]
from messages import (  # type:ignore[attr-defined]
    Hook,
    Location,
    Meta,
    ParameterType,
    Product,
    Source,
    SourceReference,
    Status,
    TestCase,
    TestCaseFinished,
    TestCaseStarted,
    TestRunFinished,
    TestRunStarted,
    TestStep,
    TestStepFinished,
    TestStepResult,
    TestStepStarted,
    Timestamp,
)
from pytest_bdd.compatibility.path import relpath
from pytest_bdd.compatibility.pytest import (
    Config,
    FixtureDef,
    FixtureRequest,
    Parser,
    get_config_root_path,
    get_metafunc_call_arg,
    is_set,
)
from pytest_bdd.packaging import get_distribution_version
from pytest_bdd.steps import StepHandler
from pytest_bdd.utils import PytestBDDIdGeneratorHandler, deepattrgetter


@attrs(eq=False)
class MessagePlugin:
    config: Config = attrib()
    current_test_case = attrib(default=None)
    current_test_case_step_to_definition_mapping = attrib(default=None)
    parameter_type_registry: Set[int] = set()
    hook_registry: Set[int] = set()

    def __attrs_post_init__(self):
        self.is_disabled = self.config.option.messages_ndjson_path is None

        if not self.is_disabled:
            self.process_messages_io_queue = Queue()
            self.process_messages_stop_event = Event()
            self.process_messages_thread = Thread(
                target=type(self).process_messages,
                args=(
                    self.process_messages_io_queue,
                    self.process_messages_stop_event,
                    self.config.option.messages_ndjson_path,
                ),
                daemon=True,
            )
            self.process_messages_thread.start()
            sleep(0)

    @staticmethod
    def process_messages(queue: Queue, stop_event: Event, messages_file_path: Union[str, Path]):
        last_enter = False
        while not (stop_event.is_set() and last_enter):  # give one more enter to take all left messages
            if stop_event.is_set():
                last_enter = True
            lock = FileLock(f"{messages_file_path}.lock")
            with lock:
                with Path(messages_file_path).open(mode="at+", buffering=1, encoding="utf-8") as f:
                    lines = []
                    while not queue.empty():
                        try:
                            message_json = queue.get(timeout=1)
                        except Empty:
                            sleep(0)
                            continue

                        try:
                            Message.model_validate(json.loads(message_json))  # type: ignore[attr-defined] # migration to pydantic2
                        except ValidationError:
                            logging.exception(f"Failed to parse:\n{pformat(message_json)}\n", exc_info=True)
                        else:
                            lines.append(f"{message_json}\n")
                        finally:
                            queue.task_done()
                        sleep(0)
                    f.writelines(lines)
                    f.flush()

    def get_timestamp(self):
        timestamp = time_ns()
        test_run_started_seconds = timestamp // 10**9
        test_run_started_nanos = timestamp - test_run_started_seconds * 10**9
        return Timestamp(seconds=test_run_started_seconds, nanos=test_run_started_nanos)

    @staticmethod
    def add_options(parser: Parser) -> None:
        """Add pytest-bdd options."""
        group = parser.getgroup("bdd", "Cucumber NDJSON")
        group.addoption(
            "--messagesndjson",
            "--messages-ndjson",
            action="store",
            dest="messages_ndjson_path",
            metavar="path",
            default=None,
            help="messages ndjson report file at given path.",
        )

    @hookimpl(hookwrapper=True)
    def pytest_generate_tests(self, metafunc):
        yield
        if self.is_disabled:
            return

        if all(
            [
                "feature" in metafunc.fixturenames,
                "scenario" in metafunc.fixturenames,
                "feature_source" in metafunc.fixturenames,
                metafunc._calls,
            ]
        ):
            config = metafunc.config

            feature_registry = set()
            pickle_registry = set()
            for call in metafunc._calls:
                feature = get_metafunc_call_arg(call, "feature")
                pickle = get_metafunc_call_arg(call, "scenario")
                feature_source: Source = get_metafunc_call_arg(call, "feature_source")

                if is_set(feature) and hasattr(feature_source, "uri") and feature_source.uri not in feature_registry:
                    feature_registry.add(feature_source.uri)
                    cast(Config, config).hook.pytest_bdd_message(config=config, message=Message(source=feature_source))

                    cast(Config, config).hook.pytest_bdd_message(
                        config=config, message=Message(gherkin_document=feature.gherkin_document)
                    )
                if is_set(pickle) and id(pickle) not in pickle_registry:
                    cast(Config, config).hook.pytest_bdd_message(config=config, message=Message(pickle=pickle))

    def pytest_bdd_message(self, config: Config, message: Message):
        if self.is_disabled:
            return

        message_json = message.model_dump_json(exclude_none=True, by_alias=True)  # type: ignore[attr-defined] # migration to pydantic2
        self.process_messages_io_queue.put_nowait(message_json)

    def pytest_runtestloop(self, session: Session):
        if self.is_disabled:
            return

        config = session.config
        hook_handler = config.hook

        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(test_run_started=TestRunStarted(timestamp=self.get_timestamp())),
        )

    def pytest_sessionstart(self, session):
        if self.is_disabled:
            return

        config = session.config
        hook_handler = config.hook

        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(
                meta=Meta(
                    protocol_version="22.0.0",
                    implementation=Product(
                        name="pytest-bdd-ng", version=str(get_distribution_version("pytest-bdd-ng"))
                    ),
                    runtime=Product(name="Python", version=sys.version),
                    os=Product(name=system(), version=version()),
                    cpu=Product(name=machine(), version=processor()),
                    ci=Ci.model_validate(obj) if (obj := detect_ci_environment(os.environ)) is not None else None,
                ),
            ),
        )

    def pytest_sessionfinish(self, session, exitstatus):
        if self.is_disabled:
            return

        config = session.config
        hook_handler = config.hook

        is_testrun_success = (isinstance(exitstatus, int) and exitstatus == 0) or exitstatus is ExitCode.OK
        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(
                test_run_finished=TestRunFinished(
                    timestamp=self.get_timestamp(),
                    success=is_testrun_success,
                )
            ),
        )
        self.process_messages_io_queue.join()
        self.process_messages_stop_event.set()
        self.process_messages_thread.join()

    @hookimpl(hookwrapper=True)
    def pytest_fixture_setup(self, fixturedef: FixtureDef, request):
        func = fixturedef.func
        func_id = id(func)

        if self.is_disabled or func_id in self.hook_registry:
            yield
            return

        config = request.config
        hook_handler = config.hook

        if hasattr(func, "__pytest_bdd_is_hook__"):
            self.hook_registry.add(func_id)

            hook_name = getattr(func, "__pytest_bdd_hook_name__", None)
            hook_expression = getattr(func, "__pytest_bdd_hook_expression__", None)

            hook_handler.pytest_bdd_message(
                config=config,
                message=Message(
                    hook=Hook(
                        id=cast(PytestBDDIdGeneratorHandler, config).pytest_bdd_id_generator.get_next_id(),
                        **({"name": hook_name} if hook_name is not None else {}),
                        source_reference=SourceReference(
                            uri=relpath(
                                getfile(func),
                                str(get_config_root_path(cast(Config, config))),
                            ),
                            location=Location(line=getsourcelines(func)[1]),
                        ),
                        **({"tag_expression": hook_expression} if hook_expression is not None else {}),
                    )
                ),
            )

        yield

    @hookimpl(hookwrapper=True)
    def pytest_runtest_setup(self, item):
        yield

        if self.is_disabled:
            return

        session = item.session
        config: Union[
            Config, PytestBDDIdGeneratorHandler
        ] = session.config  # https://github.com/python/typing/issues/213

        hook_handler = cast(Config, config).hook

        request = item._request
        scenario = request.getfixturevalue("scenario")
        feature = request.getfixturevalue("feature")

        step_registry = request.getfixturevalue("step_registry")
        seen_steps = set()
        while step_registry is not None:
            for step_definition in step_registry:
                if id(step_definition) not in seen_steps:
                    seen_steps.add(id(step_definition))
                    config.hook.pytest_bdd_message(
                        config=config, message=Message(step_definition=step_definition.as_message(config=config))
                    )
            step_registry = step_registry.parent

        step_registry = request.getfixturevalue("step_registry")
        seen_steps = set()
        while step_registry is not None:
            for step_definition in step_registry:
                if id(step_definition) not in seen_steps:
                    parameter_type_registry_getter: Callable[[FixtureRequest], ParameterTypeRegistry] = deepattrgetter(
                        "_get_parameter_type_registry", default=None
                    )(step_definition.parser)[0]

                    if parameter_type_registry_getter is None:
                        break

                    parameter_type_registry = parameter_type_registry_getter(request)

                    parameter_types = dict(
                        map(
                            lambda parameter_type: (id(parameter_type), parameter_type),
                            parameter_type_registry.parameter_types,
                        )
                    )

                    not_yet_registered_parameter_types = {
                        key: parameter_type
                        for key, parameter_type in parameter_types.items()
                        if key not in self.parameter_type_registry
                    }

                    for parameter_type in not_yet_registered_parameter_types.values():
                        hook_handler.pytest_bdd_message(
                            config=config,
                            message=Message(
                                parameter_type=ParameterType(
                                    name=parameter_type.name,
                                    regular_expressions=parameter_type.regexps,
                                    prefer_for_regular_expression_match=parameter_type._prefer_for_regexp_match,
                                    use_for_snippets=parameter_type._use_for_snippets,
                                    id=cast(PytestBDDIdGeneratorHandler, config).pytest_bdd_id_generator.get_next_id(),
                                ),
                            ),
                        )
                    self.parameter_type_registry |= not_yet_registered_parameter_types.keys()
            step_registry = step_registry.parent

        test_steps = []
        previous_step = None

        self.current_test_case_step_id_to_step_mapping = {}

        for step in scenario.steps:
            try:
                step_definition = hook_handler.pytest_bdd_match_step_definition_to_step(
                    request=request,
                    feature=feature,
                    scenario=scenario,
                    step=step,
                    previous_step=previous_step,
                )
            except StepHandler.Matcher.MatchNotFoundError as e:
                pass
            else:
                test_step = TestStep(
                    id=cast(PytestBDDIdGeneratorHandler, config).pytest_bdd_id_generator.get_next_id(),
                    pickle_step_id=step.id,
                    step_definition_ids=[step_definition.as_message(config).id]
                    # TODO Check step_match_arguments_lists
                )
                test_steps.append(test_step)
                self.current_test_case_step_id_to_step_mapping[id(step)] = test_step
            finally:
                previous_step = step

        self.current_test_case = TestCase(
            id=cast(PytestBDDIdGeneratorHandler, config).pytest_bdd_id_generator.get_next_id(),
            pickle_id=scenario.id,
            test_steps=test_steps,
        )

        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(test_case=self.current_test_case),
        )

    def pytest_bdd_before_scenario(self, request, feature, scenario):
        if self.is_disabled:
            return

        config = request.config
        hook_handler = config.hook

        self.current_test_case_start = TestCaseStarted(
            attempt=getattr(request.node, "execution_count", 0),
            id=cast(PytestBDDIdGeneratorHandler, config).pytest_bdd_id_generator.get_next_id(),
            test_case_id=self.current_test_case.id,
            worker_id=os.environ.get("PYTEST_XDIST_WORKER", "master"),
            timestamp=self.get_timestamp(),
        )

        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(test_case_started=self.current_test_case_start),
        )

    def pytest_bdd_after_scenario(self, request, feature, scenario):
        if self.is_disabled:
            return

        config = request.config

        hook_handler = config.hook

        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(
                test_case_finished=TestCaseFinished(
                    test_case_started_id=self.current_test_case_start.id,
                    timestamp=self.get_timestamp(),
                    # TODO check usage
                    will_be_retried=False,
                )
            ),
        )
        self.current_test_case = None

    def pytest_bdd_before_step(self, request, feature, scenario, step, step_func):
        if self.is_disabled:
            return

        config = request.config
        hook_handler = config.hook

        # TODO check behaviour if missing
        step_definition = self.current_test_case_step_id_to_step_mapping[id(step)]

        self.current_test_case_step_start_timestamp = self.get_timestamp()

        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(
                test_step_started=TestStepStarted(
                    test_case_started_id=self.current_test_case.id,
                    timestamp=self.current_test_case_step_start_timestamp,
                    test_step_id=step_definition.id,
                )
            ),
        )

    def pytest_bdd_after_step(self, request, feature, scenario, step, step_func):
        if self.is_disabled:
            return

        config = request.config
        hook_handler = config.hook

        # TODO check behaviour if missing
        step_definition = self.current_test_case_step_id_to_step_mapping[id(step)]
        self.current_test_case_step_finish_timestamp = self.get_timestamp()

        current_test_case_step_duration_total_nanos = (
            self.current_test_case_step_finish_timestamp.seconds * 10**9
            + self.current_test_case_step_finish_timestamp.nanos
        ) - (
            self.current_test_case_step_start_timestamp.seconds * 10**9
            + self.current_test_case_step_start_timestamp.nanos
        )
        current_test_case_step_duration_seconds = current_test_case_step_duration_total_nanos // 10**9
        current_test_case_step_duration_nanos = (
            current_test_case_step_duration_total_nanos - current_test_case_step_duration_seconds * 10**9
        )
        current_test_case_step_duration = Duration(
            seconds=current_test_case_step_duration_seconds, nanos=current_test_case_step_duration_nanos
        )

        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(
                test_step_finished=TestStepFinished(
                    test_case_started_id=self.current_test_case.id,
                    timestamp=self.current_test_case_step_finish_timestamp,
                    test_step_id=step_definition.id,
                    test_step_result=TestStepResult(duration=current_test_case_step_duration, status=Status.passed),
                )
            ),
        )

    def pytest_bdd_step_error(
        self, request, feature, scenario, step, step_func, step_func_args, exception, step_definition
    ):
        if self.is_disabled:
            return

        config = request.config
        hook_handler = config.hook

        # TODO check behaviour if missing
        step_definition = self.current_test_case_step_id_to_step_mapping[id(step)]
        self.current_test_case_step_finish_timestamp = self.get_timestamp()

        current_test_case_step_duration_total_nanos = (
            self.current_test_case_step_finish_timestamp.seconds * 10**9
            + self.current_test_case_step_finish_timestamp.nanos
        ) - (
            self.current_test_case_step_start_timestamp.seconds * 10**9
            + self.current_test_case_step_start_timestamp.nanos
        )
        current_test_case_step_duration_seconds = current_test_case_step_duration_total_nanos // 10**9
        current_test_case_step_duration_nanos = (
            current_test_case_step_duration_total_nanos - current_test_case_step_duration_seconds * 10**9
        )
        current_test_case_step_duration = Duration(
            seconds=current_test_case_step_duration_seconds, nanos=current_test_case_step_duration_nanos
        )

        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(
                test_step_finished=TestStepFinished(
                    test_case_started_id=self.current_test_case.id,
                    timestamp=self.current_test_case_step_finish_timestamp,
                    test_step_id=step_definition.id,
                    test_step_result=TestStepResult(duration=current_test_case_step_duration, status=Status.failed),
                )
            ),
        )

    def pytest_bdd_attach(self, request, attachment, media_type, file_name):
        if self.is_disabled:
            return

        config = request.config
        hook_handler = config.hook

        if isinstance(attachment, (str, TextIOBase)):
            content_encoding = ContentEncoding.identity
            _media_type = "text/plain;charset=UTF-8" if media_type is None else media_type
        elif isinstance(attachment, (bytes, bytearray, BufferedIOBase)):
            content_encoding = ContentEncoding.base64
            _media_type = "application/octet-stream" if media_type is None else media_type
        else:
            content_encoding = ContentEncoding.identity
            _media_type = "text/plain;charset=UTF-8" if media_type is None else media_type

        if isinstance(attachment, str):
            body = attachment
        elif isinstance(attachment, TextIOBase):
            body = attachment.read()
        elif isinstance(attachment, (bytes, bytearray, BufferedIOBase)):
            if isinstance(attachment, bytes):
                body_bytes = attachment
            elif isinstance(attachment, bytearray):
                body_bytes = bytes(attachment)
            elif isinstance(attachment, BufferedIOBase):
                body_bytes = attachment.read()
            else:  # pragma: no cover
                body_bytes = b""

            body = b64encode(body_bytes).decode("ascii")
        else:
            body = str(attachment)

        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(
                attachment=Attachment(
                    test_step_id=self.current_test_case.id,
                    test_case_started_id=self.current_test_case.id,
                    # TODO find a specification when it useful
                    # source=,
                    media_type=_media_type,
                    **(dict(file_name=str(file_name)) if file_name is not None else {}),
                    content_encoding=content_encoding,
                    body=body,
                )
            ),
        )



"""Feature.

The way of describing the behavior is based on Gherkin language

Syntax example:

    Feature: Articles
        Scenario: Publishing the article
            Given I'm an author user
            And I have an article
            When I go to the article page
            And I press the publish button
            Then I should not see the error message

            # Note: will query the database
            And the article should be published

:note: The "#" symbol is used for comments.
:note: There are no multiline steps, the description of the step must fit in
one line.
"""
from itertools import chain
from textwrap import dedent
from typing import Sequence, Union, cast

from attr import Factory, attrib, attrs
from gherkin.errors import CompositeParserException  # type: ignore[import]
from gherkin.pickles.compiler import Compiler  # type: ignore[import]

from messages import Background, Examples  # type:ignore[attr-defined]
from messages import Feature as FeatureMessage  # type:ignore[attr-defined]
from messages import (  # type:ignore[attr-defined]
    GherkinDocument,
    Location,
    Pickle,
    PickleStep,
    Rule,
    Scenario,
    Step,
    TableRow,
    Tag,
)
from pytest_bdd.const import TAG_PREFIX
from pytest_bdd.utils import _itemgetter, deepattrgetter


@attrs
class Feature:
    gherkin_document: GherkinDocument = attrib()
    uri = attrib()
    filename: str = attrib()

    registry: dict = attrib(default=Factory(dict))
    pickles: Sequence[Pickle] = attrib(default=Factory(list))

    def __attrs_post_init__(self):
        self.fill_registry()

    @staticmethod
    def load_pickles(scenarios_data) -> Sequence[Pickle]:
        return [*map(Pickle.model_validate, scenarios_data)]  # type: ignore[attr-defined] # migration to pydantic2

    def fill_registry(self):
        self.registry.update(self.get_child_ids_gen(self.gherkin_document.feature))

    @classmethod
    def get_child_ids_gen(cls, obj):
        if isinstance(obj, FeatureMessage):
            yield from chain.from_iterable(
                map(
                    cls.get_child_ids_gen,
                    chain(
                        obj.tags,
                        chain.from_iterable(
                            map(
                                lambda child: filter(None, [child.rule, child.background, child.scenario]), obj.children
                            )
                        ),
                    ),
                )
            )
        elif isinstance(obj, Tag):
            yield obj.id, obj
        elif isinstance(obj, Rule):
            yield obj.id, obj
            yield from chain.from_iterable(
                map(
                    cls.get_child_ids_gen,
                    chain(
                        obj.tags,
                        chain.from_iterable(
                            map(
                                lambda child: filter(
                                    None,
                                    [
                                        # TODO Check why we don't support nested Rules;
                                        # https://github.com/cucumber/gherkin/issues/126
                                        # child.rule,
                                        child.background,
                                        child.scenario,
                                    ],
                                ),
                                obj.children,
                            )
                        ),
                    ),
                )
            )
        elif isinstance(obj, Background):
            yield obj.id, obj
            yield from chain.from_iterable(map(cls.get_child_ids_gen, obj.steps))
        elif isinstance(obj, Scenario):
            yield obj.id, obj
            yield from chain.from_iterable(
                map(
                    cls.get_child_ids_gen,
                    chain(
                        obj.tags,
                        obj.steps,
                        obj.examples,
                    ),
                )
            )
        elif isinstance(obj, Examples):
            yield obj.id, obj
            yield from chain.from_iterable(
                map(cls.get_child_ids_gen, chain(obj.tags, [obj.table_header], obj.table_body))
            )
        elif isinstance(obj, TableRow):
            yield obj.id, obj
        elif isinstance(obj, Step):
            yield obj.id, obj

    load_gherkin_document = staticmethod(GherkinDocument.model_validate)  # type: ignore[attr-defined] # migration to pydantic2

    @property
    def name(self) -> Union[str, None]:
        if self.gherkin_document.feature is not None:
            return cast(str, self.gherkin_document.feature.name)
        else:
            return None

    @property
    def rel_filename(self):
        file_schema = "file"
        if self.uri.startswith(file_schema):
            return self.uri[len(file_schema) + 1 :]
        else:
            return None

    @property
    def line_number(self):
        return self.gherkin_document.feature.location.line

    @property
    def description(self):
        return dedent(self.gherkin_document.feature.description)

    @property
    def tag_names(self):
        return sorted(map(lambda tag: tag.name.lstrip(TAG_PREFIX), self.gherkin_document.feature.tags))

    def build_pickle_table_rows_breadcrumb(self, pickle):
        table_rows_lines = ",".join(
            map(
                lambda row: f"line: {deepattrgetter('location.line', default=-1)(row)[0]}",
                self._get_pickle_ast_table_rows(pickle),
            )
        )
        return f"[table_rows:[{table_rows_lines}]]" if table_rows_lines else ""

    def _get_pickle_ast_table_rows(self, pickle):
        return list(filter(lambda node: type(node) is TableRow, self._get_linked_ast_nodes(pickle)))

    def _get_linked_ast_nodes(self, obj):
        return _itemgetter(
            *((obj.ast_node_id,) if hasattr(obj, "ast_node_id") else ()),
            *getattr(obj, "ast_node_ids", ()),
        )(self.registry)

    def _get_pickle_tag_names(self, pickle: Pickle):
        return sorted(map(lambda tag: tag.name.lstrip(TAG_PREFIX), pickle.tags))  # type: ignore[no-any-return]

    def _get_pickle_ast_scenario(self, pickle: Pickle) -> Scenario:
        return cast(Scenario, next(filter(lambda node: type(node) is Scenario, self._get_linked_ast_nodes(pickle))))

    def _get_pickle_line_number(self, pickle: Pickle):
        return (
            cast(Location, location).line
            if (location := self._get_pickle_ast_scenario(pickle).location) is not None
            else -1
        )

    def _get_pickle_step_model_step(self, pickle_step: PickleStep):
        return cast(Step, next(filter(lambda node: type(node) is Step, self._get_linked_ast_nodes(pickle_step)), None))

    def _get_step_keyword(self, step: PickleStep):
        model_step: Union[Step, None] = self._get_pickle_step_model_step(step)
        if model_step is not None:
            return model_step.keyword.strip()

    def _get_step_prefix(self, step: PickleStep):
        step_keyword = self._get_step_keyword(step)
        if step_keyword is not None:
            return step_keyword.lower()

    def _get_step_line_number(self, step: PickleStep):
        model_step: Union[Step, None] = self._get_pickle_step_model_step(step)
        if model_step is not None:
            return location.line if (location := model_step.location) is not None else -1

    def _get_step_doc_string(self, step: PickleStep):
        return getattr(self._get_pickle_step_model_step(step), "doc_string", None)

    def _get_step_data_table(self, step: PickleStep):
        return getattr(self._get_pickle_step_model_step(step), "data_table", None)



import sys

if sys.version_info >= (3, 10):
    from importlib.metadata import *
    from importlib.metadata import version
else:
    from importlib_metadata import *
    from importlib_metadata import version

__all__ = ["version"]



import collections
from enum import Enum
from pathlib import Path
from typing import Callable, Iterable, Optional, Type, Union

from pytest import mark

from pytest_bdd.compatibility.parser import ParserProtocol
from pytest_bdd.compatibility.pytest import Parser
from pytest_bdd.mimetypes import Mimetype
from pytest_bdd.utils import compose, make_python_name

Args = collections.namedtuple("Args", ["args", "kwargs"])


def add_options(parser: Parser):
    """Add pytest-bdd options."""
    group = parser.getgroup("bdd", "Scenario")
    group.addoption(
        "--disable-feature-autoload",
        action="store_false",
        dest="feature_autoload",
        default=None,
        help="Turn off feature files autoload",
    )
    parser.addini(
        "disable_feature_autoload",
        default=False,
        type="bool",
        help="Turn off feature files autoload",
    )


def get_python_name_generator(name: str) -> Iterable[str]:
    """Generate a sequence of suitable python names out of given arbitrary string name."""
    python_name = make_python_name(name)
    suffix = ""
    index = 0

    def get_name() -> str:
        return "_".join(filter(bool, ["test", python_name, suffix]))

    while True:
        yield get_name()
        index += 1
        suffix = f"{index}"


test_names = get_python_name_generator("")


class FeaturePathType(Enum):
    PATH = "path"
    URL = "url"
    UNDEFINED = "undefined"


def scenario(
    feature_name: Optional[Union[Path, str]] = None,
    scenario_name: Optional[str] = None,
    encoding: str = "utf-8",
    features_base_dir: Optional[Union[Path, str]] = None,
    features_base_url=None,
    features_path_type: Optional[Union[FeaturePathType, str]] = FeaturePathType.PATH,
    features_mimetype: Optional[Mimetype] = None,
    return_test_decorator=True,
    parser_type: Optional[Type[ParserProtocol]] = None,
    parse_args=Args((), {}),
    locators=(),
):
    """
    Scenario decorator.

    :param feature_name: Feature file name. Absolute or relative to the configured feature base path.
    :param scenario_name: Scenario name.
    :param encoding: Feature file encoding.
    :param features_base_dir: Feature base directory from where features will be searched
    :param features_base_url: Feature base url from where features will be loaded
    :param features_path_type: If feature path is not absolute helps to select if filepath or url will be used
    :param features_mimetype: Helps to select appropriate parser if non-standard file extension is used
    :param return_test_decorator; Return test decorator or generated test
    :param parser_type: Parser used to parse feature-like file
    :param parse_args: args consumed by parser during parsing
    :param locators: Feature locators to load Features; Could be custom
    """
    return scenarios(
        *([feature_name] if feature_name is not None else []),
        filter_=scenario_name,
        encoding=encoding,
        features_base_dir=features_base_dir,
        features_base_url=features_base_url,
        features_path_type=features_path_type,
        features_mimetype=features_mimetype,
        return_test_decorator=return_test_decorator,
        locators=locators,
        parser_type=parser_type,
        parse_args=parse_args,
    )


def scenarios(
    *feature_paths: Union[Path, str],
    filter_: Optional[Union[str, Callable]] = None,
    return_test_decorator=False,
    encoding: str = "utf-8",
    features_base_dir: Optional[Union[Path, str]] = None,
    features_base_url: Optional[str] = None,
    features_path_type: Optional[Union[FeaturePathType, str]] = FeaturePathType.PATH,
    features_mimetype: Optional[Mimetype] = None,
    parser_type: Optional[Type[ParserProtocol]] = None,
    parse_args=Args((), {}),
    locators=(),
):
    """
    Function to bind feature files to pytest runtime

    :param feature_paths: Features file names. Absolute or relative to the configured feature base path.
    :param filter_: Callable to filter scenarios
    :param encoding: Feature file encoding.
    :param features_base_dir: Feature base directory from where features will be searched
    :param features_base_url: Feature base url from where features will be loaded
    :param features_path_type: If feature path is not absolute helps to select if filepath or url will be used
    :param features_mimetype: Helps to select appropriate parser if non-standard file extension is used
    :param return_test_decorator; Return test decorator or generated test
    :param parser_type: Parser used to parse feature-like file
    :param parse_args: args consumed by parser during parsing
    :param return_test_decorator; Return test decorator or generated test
    :param locators: Feature locators to load Features; Could be custom
    """
    if features_base_dir and features_base_url:
        raise ValueError('Both "features_base_dir" and "features_base_url" were specified')
    if features_base_dir:
        features_path_type = FeaturePathType.PATH
    elif features_base_url:
        features_path_type = FeaturePathType.URL

    decorator = compose(
        mark.pytest_bdd_scenario,
        mark.usefixtures("feature", "scenario", "feature_source"),
        mark.scenarios(
            *feature_paths,
            filter_=filter_,
            encoding=encoding,
            features_base_dir=features_base_dir,
            features_base_url=features_base_url,
            features_path_type=features_path_type,
            features_mimetype=features_mimetype,
            parser_type=parser_type,
            parse_args=parse_args,
            locators=locators,
        ),
    )

    if return_test_decorator:
        return decorator
    else:

        @decorator
        def test():
            ...

        test.__name__ = next(iter(test_names))

        return test



"""StepHandler parsers."""
from abc import ABCMeta, abstractmethod
from enum import Enum
from functools import partial, singledispatchmethod
from itertools import chain, filterfalse
from operator import attrgetter, contains, methodcaller
from re import Match
from re import Pattern as _RePattern
from re import compile as re_compile
from typing import Any, Collection, Dict, Iterable, Optional, Protocol, Sequence, Type, Union, cast, runtime_checkable

import parse as base_parse
import parse_type.cfparse as base_cfparse
from cucumber_expressions.argument import Argument as CucumberExpressionArgument
from cucumber_expressions.errors import CantEscape, UndefinedParameterTypeError
from cucumber_expressions.expression import CucumberExpression
from cucumber_expressions.parameter_type_registry import ParameterTypeRegistry
from cucumber_expressions.regular_expression import RegularExpression as CucumberRegularExpression

from messages import ExpressionType  # type:ignore[attr-defined]
from pytest_bdd.compatibility.pytest import FixtureRequest
from pytest_bdd.model.messages_extension import ExpressionType as ExpressionTypeExtension
from pytest_bdd.utils import StringableProtocol, stringify


class ParserBuildValueError(ValueError):
    ...


@runtime_checkable
class StepParserProtocol(Protocol):
    type: Union[ExpressionType, ExpressionTypeExtension, str] = ExpressionTypeExtension.pytest_bdd_other_expression

    def parse_arguments(
        self, request: FixtureRequest, name: str, anonymous_group_names: Optional[Iterable[str]] = None
    ) -> Optional[Dict[str, Any]]:
        ...  # pragma: no cover

    @property
    def arguments(self) -> Collection[str]:
        ...  # pragma: no cover

    def is_matching(self, request: FixtureRequest, name: str) -> bool:
        ...  # pragma: no cover

    def __str__(self) -> str:
        ...  # pragma: no cover


class RegistryMode(Enum):
    NEW = "NEW"
    GLOBAL = "GLOBAL"
    FIXTURE = "FIXTURE"
    NOT_DEFINED = None


class StepParser(StepParserProtocol, metaclass=ABCMeta):
    """Parser of the individual step."""

    @abstractmethod
    def parse_arguments(
        self, request: FixtureRequest, name: str, anonymous_group_names: Optional[Iterable[str]] = None
    ) -> Optional[Dict[str, Any]]:
        """Get step arguments from the given step name.

        :return: `dict` of step arguments
        """
        raise NotImplementedError()  # pragma: no cover

    @property
    @abstractmethod
    def arguments(self) -> Collection[str]:
        """Get step argument names from the given step name."""
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def is_matching(self, request: FixtureRequest, name: str) -> bool:
        """Match given name with the step name."""
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def __str__(self) -> str:
        """Match given name with the step name."""
        raise NotImplementedError()  # pragma: no cover

    @classmethod
    def build(cls, parserlike: Union[str, bytes, "StepParser", StepParserProtocol]) -> "StepParser":
        """Get parser by given name.

        :param parserlike: name of the step to parse

        :return: step parser object
        :rtype: StepParser
        """

        if isinstance(parserlike, StepParserProtocol):
            parser = cast(StepParser, parserlike)
        elif isinstance(parserlike, _RePattern):
            parser = re(parserlike)
        elif isinstance(parserlike, base_parse.Parser):
            parser = parse(parserlike)
        elif isinstance(parserlike, CucumberExpression):
            parser = cucumber_expression(parserlike)
        elif isinstance(parserlike, CucumberRegularExpression):
            parser = cucumber_regular_expression(parserlike)
        else:
            parser = heuristic(parserlike)

        return parser


class re(StepParser):
    """Regex step parser."""

    type = ExpressionTypeExtension.pytest_bdd_regular_expression

    # https://bugs.python.org/issue45684
    @singledispatchmethod  # type:ignore[misc]
    def __init__(self, *args, **kwargs):
        raise NotImplementedError()  # pragma: no cover

    @__init__.register
    def _(self, pattern: str, *args: Any, **kwargs: Any) -> None:
        """Compile regex."""
        self.pattern = pattern
        self.regex = re_compile(self.pattern, *args, **kwargs)

    @__init__.register
    def _(self, pattern: _RePattern):
        """Compile regex."""
        self.pattern = pattern.pattern
        self.regex = pattern

    def parse_arguments(
        self,
        request: FixtureRequest,
        name,
        anonymous_group_names: Optional[Iterable[str]] = None,
    ):
        match = cast(Match, self.regex.fullmatch(name))  # Can't be None because is already matched
        group_dict = match.groupdict()
        if anonymous_group_names is not None:
            group_dict.update(
                zip(
                    anonymous_group_names,
                    map(
                        lambda span: name[slice(*span)],  # type: ignore[no-any-return] # https://github.com/python/mypy/issues/9590
                        filterfalse(
                            partial(contains, [*map(match.span, group_dict.keys())]),
                            map(match.span, range(1, len(match.groups()) + 1)),
                        ),
                    ),
                )
            )

        return {k: v for k, v in group_dict.items() if v is not None}

    @property
    def arguments(self):
        return [*self.regex.groupindex.keys()]

    def is_matching(self, request: FixtureRequest, name):
        return bool(self.regex.fullmatch(name))

    def __str__(self):
        return stringify(self.pattern)


class parse(StepParser):
    """parse step parser."""

    type = ExpressionTypeExtension.pytest_bdd_parse_expression

    # https://bugs.python.org/issue45684
    @singledispatchmethod  # type:ignore[misc]
    def __init__(self, format, *args, **kwargs):
        if isinstance(format, (StringableProtocol, str, bytes)):
            self.__init_stringable__(format, *args, **kwargs)
        else:
            raise ParserBuildValueError(f"Unable build parser for format {format}")  # pragma: no cover

    def __init_stringable__(
        self, format: Union[StringableProtocol, str, bytes], *args: Any, builder=base_parse.compile, **kwargs: Any
    ) -> None:
        self.format = stringify(format)
        self.parser = builder(self.format, *args, **kwargs)

    @__init__.register
    def _(self, format: base_parse.Parser):
        self.format = format._format
        self.parser = format

    @classmethod
    def cfparse(cls, *args, **kwargs):
        kwargs.setdefault("builder", base_cfparse.Parser)
        return cls(*args, **kwargs)

    def parse_arguments(
        self, request: FixtureRequest, name: str, anonymous_group_names: Optional[Iterable[str]] = None
    ) -> Union[Dict[str, Any]]:
        match = self.parser.parse(name)
        group_dict = cast(dict, match.named)
        if anonymous_group_names is not None:
            group_dict.update(dict(zip(anonymous_group_names, match.fixed)))
        return group_dict

    @property
    def arguments(self) -> Collection[str]:
        return [*self.parser._match_re.groupindex.keys()]

    def is_matching(self, request: FixtureRequest, name):
        try:
            return bool(self.parser.parse(name))
        except ValueError:
            return False

    def __str__(self):
        return str(self.format)


class cfparse(parse):
    """cfparse step parser."""

    type = ExpressionTypeExtension.pytest_bdd_cfparse_expression

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("builder", base_cfparse.Parser)
        super().__init__(*args, **kwargs)


class string(StepParser):
    """Exact string step parser."""

    type = ExpressionTypeExtension.pytest_bdd_string_expression

    def __init__(self, name: Union[StringableProtocol, str, bytes]) -> None:
        self.name = stringify(name)

    def parse_arguments(
        self, request: FixtureRequest, name: str, anonymous_group_names: Optional[Iterable[str]] = None
    ) -> Dict[str, Any]:
        """No parameters are available for simple string step.

        :return: `dict` of step arguments
        """
        return {}

    @property
    def arguments(self):
        return []

    def is_matching(self, request: FixtureRequest, name: str) -> bool:
        """Match given name with the step name."""
        return bool(self.name == name)

    def __str__(self):
        return self.name


@runtime_checkable
class _CucumberExpressionProtocol(Protocol):
    def match(self, text: str) -> Optional[Sequence[CucumberExpressionArgument]]:
        ...  # pragma: no cover


class _CucumberExpression(StepParser):
    pattern: str

    expression_type: Type[Union[CucumberExpression, CucumberRegularExpression]]
    parameter_type_registry_like: Union[ParameterTypeRegistry, Any]
    parameter_type_registry = ParameterTypeRegistry()  # default registry

    def is_matching(self, request: FixtureRequest, name: str) -> bool:
        try:
            return bool(self.rebuild_expression_in_test_context(request).tree_regexp.match(name))
        except (UndefinedParameterTypeError, CantEscape):
            return False

    def parse_arguments(
        self, request: FixtureRequest, name: str, anonymous_group_names: Optional[Iterable[str]] = None
    ) -> Optional[Dict[str, Any]]:
        return dict(
            zip(
                anonymous_group_names or [],
                map(attrgetter("value"), self.rebuild_expression_in_test_context(request).match(name) or []),
            )
        )

    def __str__(self):
        return str(self.pattern)

    def rebuild_expression_in_test_context(self, request) -> Union[CucumberExpression, CucumberRegularExpression]:
        return self.expression_type(self.pattern, self._get_parameter_type_registry(request))

    def _get_parameter_type_registry(self, request) -> Union[ParameterTypeRegistry, Any]:
        if (
            isinstance(self.parameter_type_registry_like, (str, RegistryMode))
            or self.parameter_type_registry_like is None
        ):
            parameter_type_registry_mode = RegistryMode(self.parameter_type_registry_like)

            parameter_type_registry = {
                RegistryMode.NEW: ParameterTypeRegistry,
                RegistryMode.GLOBAL: lambda: self.parameter_type_registry,
                RegistryMode.NOT_DEFINED: lambda: self.parameter_type_registry,
                RegistryMode.FIXTURE: lambda: request.getfixturevalue("parameter_type_registry"),
            }[parameter_type_registry_mode]()
        else:
            parameter_type_registry = self.parameter_type_registry_like
        return parameter_type_registry


class cucumber_expression(_CucumberExpression):
    type = ExpressionType.cucumber_expression
    expression_type = CucumberExpression

    # https://bugs.python.org/issue45684
    @singledispatchmethod  # type:ignore[misc]
    def __init__(self, *args, **kwargs):
        raise NotImplementedError()  # pragma: no cover

    @__init__.register
    def _(
        self,
        expression: str,
        parameter_type_registry: Union[ParameterTypeRegistry, RegistryMode, Any] = RegistryMode.FIXTURE,
    ):
        self.pattern = expression
        self.parameter_type_registry_like = parameter_type_registry

    @__init__.register
    def _(
        self,
        expression: CucumberExpression,
    ):
        self.pattern = expression.expression
        self.parameter_type_registry_like = expression.parameter_type_registry

    @property
    def arguments(self):
        return []


class cucumber_regular_expression(_CucumberExpression):
    type = ExpressionType.regular_expression
    expression_type = CucumberRegularExpression
    # https://bugs.python.org/issue45684

    @singledispatchmethod  # type:ignore[misc]
    def __init__(self, *args, **kwargs):
        raise NotImplementedError()  # pragma: no cover

    @__init__.register
    def _(
        self,
        expression: str,
        parameter_type_registry: Union[ParameterTypeRegistry, RegistryMode, Any] = RegistryMode.FIXTURE,
    ):
        self.pattern = expression
        self.parameter_type_registry_like = parameter_type_registry

    @__init__.register
    def _(
        self,
        expression: CucumberRegularExpression,
    ):
        self.pattern = expression.expression_regexp.pattern
        self.parameter_type_registry_like = expression.parameter_type_registry

    @property
    def arguments(self) -> Collection[str]:
        return [*re_compile(self.pattern).groupindex.keys()]


class heuristic(StepParser):
    type = ExpressionTypeExtension.pytest_bdd_heuristic_expression

    def __init__(
        self,
        format,
        parameter_type_registry: Optional[Union[ParameterTypeRegistry, RegistryMode, Any]] = RegistryMode.FIXTURE,
    ):
        if isinstance(format, (StringableProtocol, str, bytes)):
            self.format = stringify(format)
        else:
            self.format = format
        self.parameter_type_registry = parameter_type_registry
        self.parsers_are_built = False
        self.build_parsers()

    def build_parsers(self):
        if self.parsers_are_built:
            return

        # Rework to exception groups after python 3.10 end of support
        e_cause = None
        try:
            self.string_parser: Optional[string] = string(self.format)
        except Exception as e:
            e_cause = e
            self.string_parser = None
        try:
            self.cucumber_expression_parser = cucumber_expression(
                self.format, parameter_type_registry=self.parameter_type_registry
            )
        except Exception as e:
            e.__cause__, e_cause = e_cause, e
            self.cucumber_expression_parser = None

        try:
            self.cfparse_parser: Optional[cfparse] = cfparse(self.format)
        except Exception as e:
            e.__cause__, e_cause = e_cause, e
            self.cfparse_parser = None

        try:
            self.re_parser = re(self.format)
        except Exception as e:
            e.__cause__, e_cause = e_cause, e
            self.re_parser = None

        self.parsers_are_built = True
        if not any(self.parser_by_priorities):
            raise ParserBuildValueError(
                f"Unable build parser for format {self.format}"
            ) from e_cause  # pragma: no cover

    @property
    def parser_by_priorities(self) -> Sequence[Optional[StepParser]]:
        return [self.string_parser, self.cucumber_expression_parser, self.cfparse_parser, self.re_parser]

    def is_matching(self, request: FixtureRequest, name: str) -> bool:
        return any(map(methodcaller("is_matching", request, name), filter(bool, self.parser_by_priorities)))

    def parse_arguments(
        self, request: FixtureRequest, name: str, anonymous_group_names: Optional[Iterable[str]] = None
    ) -> Optional[Dict[str, Any]]:
        for parser in self.parser_by_priorities:
            if parser is not None and parser.is_matching(request, name):
                arguments = parser.parse_arguments(request, name, anonymous_group_names=anonymous_group_names)
                break
        else:
            arguments = None
        return arguments

    @property
    def arguments(self) -> Collection[str]:
        return [
            *chain.from_iterable(
                map(
                    lambda parser: []  # type:ignore[no-any-return]
                    if (args := getattr(parser, "arguments")) is None
                    else args,
                    self.parser_by_priorities,
                )
            )
        ]

    def __str__(self):
        return self.format



from itertools import cycle
from operator import attrgetter
from typing import AbstractSet, List, Optional, Protocol, Type, TypeVar, Union, runtime_checkable

from _pytest.mark import Mark
from attr import attrib, attrs
from cucumber_tag_expressions import TagExpressionError, TagExpressionParser

from pytest_bdd.compatibility.pytest import PYTEST6, PYTEST83

if PYTEST6:
    from pytest_bdd.compatibility.pytest import Expression, MarkMatcher, ParseError

TagExpressionType = TypeVar("TagExpressionType", bound="TagExpression")


@runtime_checkable
class TagExpression(Protocol):
    @classmethod
    def parse(cls: Type[TagExpressionType], expression: str) -> TagExpressionType:
        raise NotImplementedError  # pragma: no cover

    def evaluate(self, marks: List[Mark]) -> bool:
        raise NotImplementedError  # pragma: no cover


@attrs
class _ModernTagExpression(TagExpression):
    expression: Optional["Expression"] = attrib()

    @classmethod
    def parse(cls, expression):
        try:
            return cls(expression=Expression.compile(expression) if expression != "" else None)
        except ParseError as e:
            raise ValueError(f"Unable parse mark expression: {expression}: {e}") from e


@attrs
class _EnhancedMarksTagExpression(_ModernTagExpression):
    """Used for 8.3<=pytest"""

    def evaluate(self, marks):
        return self.expression.evaluate(MarkMatcher.from_markers(marks)) if self.expression is not None else True


@attrs
class _MarksTagExpression(_ModernTagExpression):
    """Used for 6.0<=pytest<8.3"""

    def evaluate(self, marks):
        return (
            self.expression.evaluate(MarkMatcher(map(attrgetter("name"), marks)))
            if self.expression is not None
            else True
        )


@attrs
class _FallbackMarksTagExpression(TagExpression):
    """Used for pytest<6.0"""

    expression: Optional[str] = attrib()

    @classmethod
    def parse(cls, expression):
        try:
            if expression != "":
                eval(expression, {})
        except SyntaxError as e:
            raise ValueError(f"Unable parse mark expression: {expression}: {e}") from e
        except NameError:
            pass
        return cls(expression=expression if expression != "" else None)

    def evaluate(self, marks):
        return (
            eval(self.expression, {}, dict(zip(map(attrgetter("name"), marks), cycle([True]))))
            if self.expression is not None
            else True
        )


MarksTagExpression: Type[Union[_EnhancedMarksTagExpression, _MarksTagExpression, _FallbackMarksTagExpression]]
if PYTEST83:
    MarksTagExpression = _EnhancedMarksTagExpression
elif PYTEST6:
    MarksTagExpression = _MarksTagExpression
else:
    MarksTagExpression = _FallbackMarksTagExpression


@attrs
class GherkinTagExpression(TagExpression):
    expression: TagExpressionParser = attrib()

    @classmethod
    def parse(cls, expression):
        try:
            return cls(expression=TagExpressionParser.parse(expression))
        except TagExpressionError as e:
            raise ValueError(f"Unable parse tag expression: {expression}: {e}") from e

    def evaluate(self, marks):
        return self.expression.evaluate(map(attrgetter("name"), marks))



"""pytest-bdd missing test code generation."""
import argparse
import os.path
from itertools import chain, filterfalse, zip_longest
from operator import lt, methodcaller
from pathlib import Path
from typing import Iterable, Optional, Sequence, Tuple, Union, cast

import py
from mako.template import Template

from messages import Pickle, PickleStep  # type:ignore[attr-defined]
from pytest_bdd.compatibility.importlib.resources import as_file, files
from pytest_bdd.compatibility.pytest import Config, ExitCode, FixtureRequest, Item, Parser, Session, wrap_session
from pytest_bdd.model import Feature, StepType
from pytest_bdd.packaging import compare_distribution_version
from pytest_bdd.parser import GherkinParser
from pytest_bdd.steps import StepHandler
from pytest_bdd.utils import make_python_name

STEP_TYPE_TO_STEP_PREFIX = {
    StepType.unknown: "*",
    StepType.outcome: "Then",
    StepType.context: "Given",
    StepType.action: "When",
}


STEP_TYPE_TO_STEP_METHOD_NAME = {
    StepType.unknown: "step",
    StepType.outcome: "then",
    StepType.context: "given",
    StepType.action: "when",
}


def check_existense(file_name):
    """Check file or directory name for existence."""
    if not os.path.exists(file_name):
        raise argparse.ArgumentTypeError(f"{file_name} is an invalid file or directory name")
    return Path(file_name)


def add_options(parser: Parser) -> None:
    """Add pytest-bdd options."""
    group = parser.getgroup("bdd", "Generation")

    group._addoption(
        "--generate-missing",
        action="store_true",
        dest="generate_missing",
        default=False,
        help="Generate missing bdd test code for given feature files and exit.",
    )

    group._addoption(
        "--generate",
        action="store_true",
        dest="generate",
        default=False,
        help="Generate bdd test code for given feature files and exit.",
    )

    group._addoption(
        "--feature",
        metavar="FILE_OR_DIR",
        action="append",
        type=check_existense,
        dest="features",
        help="Feature file or directory to generate code for. Multiple allowed.",
    )


def cmdline_main(config: Config) -> Optional[int]:
    """Check config option to show missing code."""
    if config.option.generate_missing:
        return generate_and_print_missing_code(config)
    elif config.option.generate:
        return generate_and_print_code(config)
    else:
        return None  # Make mypy happy


def generate_code(
    features: Sequence[Feature],
    feature_pickles: Sequence[Tuple[Feature, Pickle]],
    feature_pickle_steps: Sequence[Tuple[Tuple[Feature, Pickle], PickleStep]],
) -> str:
    """Generate test code for the given filenames."""
    with as_file(files("pytest_bdd.template").joinpath("test.py.mak")) as path:
        template = Template(filename=str(path))
    code = template.render(
        features=features,
        feature_pickles=feature_pickles,
        feature_pickle_steps=feature_pickle_steps,
        make_python_name=make_python_name,
        make_python_docstring=make_python_docstring,
        make_string_literal=make_string_literal,
        step_type_to_method_name=STEP_TYPE_TO_STEP_METHOD_NAME,
    )
    return cast(str, code)


def generate_and_print_missing_code(config: Config) -> Union[int, ExitCode]:
    """Wrap pytest session to show missing code."""

    def _(config: Config, session: Session) -> None:
        """Preparing fixture duplicates for output."""
        tw = py.io.TerminalWriter()
        config.hook.pytest_collection(session=session)

        if config.option.features is None:
            tw.line("The --feature parameter is required.", red=True)
            session.exitstatus = 100
            return

        seen_feature_pickles_ids = set()
        non_matched_feature_pickle_steps = []

        for item in session.items:
            is_legacy_pytest = compare_distribution_version("pytest", "7.0", lt)

            method_name = "prepare" if is_legacy_pytest else "setup"
            methodcaller(method_name, item)(item.session._setupstate)

            item = cast(Item, item)
            item_request: FixtureRequest = item._request
            pickle: Pickle = item_request.getfixturevalue("scenario")
            feature: Feature = item_request.getfixturevalue("feature")

            seen_feature_pickles_ids.add((feature.uri, pickle.name))

            previous_step = None
            for step in pickle.steps:
                try:
                    item_request.config.hook.pytest_bdd_match_step_definition_to_step(
                        request=item_request, feature=feature, scenario=pickle, step=step, previous_step=previous_step
                    )
                except StepHandler.Matcher.MatchNotFoundError:
                    non_matched_feature_pickle_steps.append(((feature, pickle), step))
                finally:
                    previous_step = step

            item.session._setupstate.teardown_exact(  # type: ignore[call-arg]
                *((item,) if is_legacy_pytest else ()), None
            )

        features = GherkinParser().get_from_paths(config, list(map(Path, config.option.features)))

        seen_features_uris = set()
        for feature_uri, pickle_name in seen_feature_pickles_ids:
            if feature_uri not in seen_features_uris:
                seen_features_uris.add(feature_uri)

        non_seen_features = list(filterfalse(lambda feature: feature.uri in seen_features_uris, features))

        non_seen_feature_pickles = list(
            filter(
                lambda feature_pickle: (feature_pickle[0].uri, feature_pickle[1].name) not in seen_feature_pickles_ids,
                chain.from_iterable(map(lambda feature: zip_longest((), feature.pickles, fillvalue=feature), features)),
            )
        )

        unique_step_defs_ids = {(step.type, step.text) for _, step in non_matched_feature_pickle_steps}
        unique_non_matched_feature_pickle_steps = list(
            map(
                lambda step_def_id: next(
                    filter(
                        lambda feature_pickle_step: (
                            feature_pickle_step[1].type == step_def_id[0]
                            and feature_pickle_step[1].text == step_def_id[1]
                        ),
                        non_matched_feature_pickle_steps,
                    )
                ),
                unique_step_defs_ids,
            )
        )

        print_missing_code(
            non_seen_features,
            non_seen_feature_pickles,  # type: ignore[arg-type]
            non_matched_feature_pickle_steps,
            unique_non_matched_feature_pickle_steps,
        )

        if non_seen_feature_pickles or non_matched_feature_pickle_steps:
            session.exitstatus = 100

    return wrap_session(config=config, doit=_)


def generate_and_print_code(config: Config) -> Union[int, ExitCode]:
    """Wrap pytest session to show missing code."""

    def _(config: Config, session: Session) -> None:
        """Preparing fixture duplicates for output."""
        tw = py.io.TerminalWriter()

        if config.option.features is None:
            tw.line("The --feature parameter is required.", red=True)
            session.exitstatus = 100
            return

        features = GherkinParser().get_from_paths(config, list(map(Path, config.option.features)))

        feature_pickles: Sequence[Tuple[Feature, Pickle]] = list(
            chain.from_iterable(
                map(
                    lambda feature: cast(
                        Iterable[Tuple[Feature, Pickle]], zip_longest((), feature.pickles, fillvalue=feature)
                    ),
                    features,
                )
            )
        )

        feature_pickles_steps: Sequence[Tuple[Tuple[Feature, Pickle], PickleStep]] = list(
            chain.from_iterable(
                map(
                    lambda feature_pickle: cast(
                        Iterable[Tuple[Tuple[Feature, Pickle], PickleStep]],
                        zip_longest((), feature_pickle[1].steps, fillvalue=feature_pickle),
                    ),
                    feature_pickles,
                )
            )
        )

        unique_step_defs_ids = {(step.type, step.text) for (feature, pickle), step in feature_pickles_steps}
        unique_feature_pickle_steps = sorted(
            list(
                map(
                    lambda step_def_id: next(  # type: ignore[no-any-return]
                        filter(
                            lambda s: (s[1].type == step_def_id[0] and s[1].text == step_def_id[1]),
                            feature_pickles_steps,
                        )
                    ),
                    unique_step_defs_ids,
                )
            ),
            key=lambda feature_pickle_step: cast(str, feature_pickle_step[1].text),
        )

        code = generate_code(features, feature_pickles, unique_feature_pickle_steps)
        tw.write(code)

    verbosity = config.option.verbose
    try:
        config.option.verbose = -2
        exit_code = wrap_session(config=config, doit=_)
    finally:
        config.option.verbose = verbosity

    return exit_code


def print_missing_code(
    features,
    feature_pickles: Sequence[Tuple[Feature, Pickle]],
    feature_pickle_steps: Sequence[Tuple[Tuple[Feature, Pickle], PickleStep]],
    unique_steps,
) -> None:
    """Print missing code with TerminalWriter."""
    tw = py.io.TerminalWriter()
    scenario = step = None

    for feature, pickle in feature_pickles:
        tw.line()
        tw.line(
            f'Scenario "{pickle.name}" is not bound to any test in the feature "{feature.name}"'
            f" in the file {feature.filename}:{feature._get_pickle_line_number(pickle)}",
            red=True,
        )

    if scenario:
        tw.sep("-", red=True)

    for (feature, pickle), step in feature_pickle_steps:
        tw.line()
        step_type = STEP_TYPE_TO_STEP_PREFIX[step.type if step.type is not None else StepType.unknown]
        tw.line(
            f"""StepHandler {step_type} "{step.text}" is not defined in the scenario "{pickle.name}" in the feature"""
            f""" "{feature.name}" in the file"""
            f""" {feature.filename}:{feature._get_step_line_number(step)}""",
            red=True,
        )

    if step:
        tw.sep("-", red=True)

    tw.line("Please place the code above to the test file(s):")
    tw.line()

    code = generate_code(features, feature_pickles, unique_steps)
    tw.write(code)


def make_python_docstring(string: str) -> str:
    """Make a python docstring literal out of a given string."""
    return '"""{}."""'.format(string.replace('"""', '\\"\\"\\"'))


def make_string_literal(string: str) -> str:
    """Make python string literal out of a given string."""
    return "'{}'".format(string.replace("'", "\\'"))



from collections import deque
from contextlib import suppress
from functools import partial
from inspect import signature
from itertools import chain, starmap
from operator import attrgetter, contains, methodcaller
from pathlib import Path
from types import ModuleType
from typing import Any, Collection, Deque, Iterable, Optional, Sequence, Union
from unittest.mock import patch

import pytest
from _pytest.nodes import Collector
from pathvalidate import is_valid_filepath

from messages import Pickle  # type:ignore[attr-defined]
from messages import PickleStep as Step  # type:ignore[attr-defined]
from pytest_bdd import cucumber_json, generation, gherkin_terminal_reporter, given, steps, then, when
from pytest_bdd.allure_logging import AllurePytestBDD
from pytest_bdd.collector import FeatureFileModule as FeatureFileCollector
from pytest_bdd.collector import Module as ModuleCollector
from pytest_bdd.compatibility.pytest import (
    PYTEST7,
    Config,
    FixtureRequest,
    Mark,
    MarkDecorator,
    Metafunc,
    Parser,
    PytestPluginManager,
)
from pytest_bdd.compatibility.struct_bdd import STRUCT_BDD_INSTALLED
from pytest_bdd.message_plugin import MessagePlugin
from pytest_bdd.mimetypes import Mimetype
from pytest_bdd.model import Feature
from pytest_bdd.parser import GherkinParser
from pytest_bdd.parsers import cucumber_expression
from pytest_bdd.reporting import ScenarioReporterPlugin
from pytest_bdd.runner import ScenarioRunner
from pytest_bdd.scenario import FeaturePathType
from pytest_bdd.scenario import add_options as scenario_add_options
from pytest_bdd.scenario import scenarios
from pytest_bdd.scenario_locator import FileScenarioLocator, UrlScenarioLocator
from pytest_bdd.steps import StepHandler
from pytest_bdd.utils import IdGenerator, compose, getitemdefault, is_url_parsable, setdefaultattr

if STRUCT_BDD_INSTALLED:
    from pytest_bdd.struct_bdd.plugin import StructBDDPlugin


def pytest_addhooks(pluginmanager: PytestPluginManager) -> None:
    """Register plugin hooks."""
    from pytest_bdd import hooks

    pluginmanager.add_hookspecs(hooks)


@given("trace")
@when("trace")
@then("trace")
def trace() -> None:
    """Enter pytest's pdb trace."""
    pytest.set_trace()


__registry = StepHandler.Registry()


@pytest.fixture
def step_registry() -> StepHandler.Registry:
    """Fixture containing registry of all user-defined steps"""
    return __registry


step_registry.__pytest_bdd_step_registry__ = __registry  # type: ignore[attr-defined]


@pytest.fixture
def step_matcher(pytestconfig) -> StepHandler.Matcher:
    """Fixture containing matcher to help find step definition for selected step of scenario"""
    return StepHandler.Matcher(pytestconfig)  # type: ignore[call-arg]


@pytest.fixture
def steps_left() -> Deque[Step]:
    """Fixture containing steps which are left to be executed"""
    return deque()


@pytest.fixture
def parameter_type_registry():
    """Fixture parameter type registry for Cucumber expressions"""
    return cucumber_expression.parameter_type_registry


@pytest.fixture
def attach(request: FixtureRequest):
    """Fixture parameter type registry for Cucumber expressions"""

    def add_attachment(attachment, media_type: Optional[str] = None, file_name=None):
        request.config.hook.pytest_bdd_attach(
            request=request, attachment=attachment, media_type=media_type, file_name=file_name
        )

    return add_attachment


def pytest_addoption(parser: Parser) -> None:
    """Add pytest-bdd options."""
    add_bdd_ini(parser)
    steps.add_options(parser)
    scenario_add_options(parser)
    cucumber_json.add_options(parser)
    generation.add_options(parser)
    gherkin_terminal_reporter.add_options(parser)
    MessagePlugin.add_options(parser)


def add_bdd_ini(parser: Parser) -> None:
    parser.addini("bdd_features_base_dir", "Base features directory.")
    parser.addini("bdd_features_base_url", "Base features url.")


@pytest.mark.trylast
def pytest_configure(config: Config) -> None:
    """Configure all subplugins."""
    config.addinivalue_line("markers", "pytest_bdd_scenario: marker to identify pytest_bdd tests")
    config.addinivalue_line("markers", "scenarios: marker to provide scenarios locator")
    cucumber_json.configure(config)
    gherkin_terminal_reporter.configure(config)
    config.pluginmanager.register(ScenarioReporterPlugin())
    config.pluginmanager.register(ScenarioRunner())
    config.pluginmanager.register(MessagePlugin(config=config), name="pytest_bdd_messages")  # type: ignore[call-arg]
    config.__allure_plugin__ = AllurePytestBDD.register_if_allure_accessible(config)  # type: ignore[attr-defined]
    setdefaultattr(config, "pytest_bdd_id_generator", value_factory=IdGenerator)
    if STRUCT_BDD_INSTALLED:
        config.pluginmanager.register(StructBDDPlugin())


@pytest.hookimpl(tryfirst=True)
def pytest_unconfigure(config: Config) -> None:
    config.pluginmanager.unregister(name="pytest_bdd_messages")
    with suppress(AttributeError):
        config.__allure_plugin__.unregister(config)  # type: ignore[attr-defined]
    cucumber_json.unconfigure(config)


def _pytest_pycollect_makemodule():
    with patch("_pytest.python.Module", new=ModuleCollector):
        yield


if PYTEST7:

    @pytest.hookimpl(hookwrapper=True)
    def pytest_pycollect_makemodule(parent, module_path):
        yield from _pytest_pycollect_makemodule()

else:

    @pytest.hookimpl(hookwrapper=True)
    def pytest_pycollect_makemodule(path, parent):
        yield from _pytest_pycollect_makemodule()


@pytest.hookimpl(tryfirst=True)
def pytest_plugin_registered(plugin, manager):
    if hasattr(plugin, "__file__") and isinstance(plugin, (type, ModuleType)):
        StepHandler.Registry.inject_registry_fixture_and_register_steps(plugin)


def _build_filter(filter_):
    if callable(filter_):
        updated_filter = filter_
    else:
        if filter_ is None:
            updated_filter = None
        else:
            if not isinstance(filter_, str):
                filter_ = str(filter_)

            def updated_filter(config, feature, scenario):
                return scenario.name == filter_

    return updated_filter


def _build_scenario_locators_from_mark(mark: Mark, config: Config) -> Iterable[Any]:
    raw_mark_arguments = signature(scenarios).bind(*mark.args, **mark.kwargs)
    raw_mark_arguments.apply_defaults()
    mark_arguments = raw_mark_arguments.arguments

    locators_iterables = [mark_arguments["locators"]]

    filter_ = _build_filter(getitemdefault(mark_arguments, "filter_", default=None))

    features_base_dir = mark.kwargs.get("features_base_dir")
    if features_base_dir is None:
        try:
            features_base_dir = config.getini("bdd_features_base_dir")
        except (ValueError, KeyError):
            features_base_dir = config.rootpath
    if callable(features_base_dir):
        features_base_dir = features_base_dir(config)

    features_base_url = mark.kwargs.get("features_base_url")
    if features_base_url is None:
        try:
            features_base_url = config.getini("bdd_features_base_url") or None
        except (ValueError, KeyError):
            features_base_url = None
    if callable(features_base_url):
        features_base_url = features_base_url(config)

    features_path_type = mark_arguments["features_path_type"]
    if features_path_type is None:
        features_path_type = FeaturePathType.UNDEFINED
    elif isinstance(features_path_type, str):
        features_path_type = FeaturePathType(features_path_type)
    elif isinstance(features_path_type, FeaturePathType):
        pass
    else:
        raise ValueError(f"Unknown feature path type")

    feature_paths = list(mark_arguments["feature_paths"] or [])

    if features_path_type is FeaturePathType.PATH:
        file_locator_feature_paths = feature_paths
    elif features_path_type is FeaturePathType.UNDEFINED:
        file_locator_feature_paths = [*filter(lambda p: is_valid_filepath(Path(p), platform="auto"), feature_paths)]
    else:
        file_locator_feature_paths = []

    path_locator = FileScenarioLocator(  # type: ignore[call-arg]
        feature_paths=file_locator_feature_paths,
        filter_=filter_,
        encoding=mark_arguments["encoding"],
        features_base_dir=features_base_dir,
        mimetype=mark_arguments["features_mimetype"],
        parser_type=mark_arguments["parser_type"],
        parse_args=mark_arguments["parse_args"],
    )

    locators_iterables.append([path_locator])

    if features_path_type is FeaturePathType.URL:
        url_locator_feature_paths = feature_paths
    elif features_path_type is FeaturePathType.UNDEFINED:
        url_locator_feature_paths = [*filter(is_url_parsable, feature_paths)]
    else:
        url_locator_feature_paths = []

    url_locator = UrlScenarioLocator(  # type: ignore[call-arg]
        url_paths=url_locator_feature_paths,
        filter_=mark_arguments["filter_"],
        encoding=mark_arguments["encoding"],
        features_base_url=features_base_url,
        mimetype=mark_arguments["features_mimetype"],
        parser_type=mark_arguments["parser_type"],
        parse_args=mark_arguments["parse_args"],
    )

    locators_iterables.append([url_locator])
    return chain(*locators_iterables)


def _build_scenario_param(feature: Feature, pickle: Pickle, feature_data: str, config: Config):
    marks = []
    for tag in feature._get_pickle_tag_names(pickle):
        tag_marks = config.hook.pytest_bdd_convert_tag_to_marks(feature=feature, scenario=pickle, tag=tag)
        if tag_marks is not None:
            marks.extend(tag_marks)
    return pytest.param(
        feature,
        pickle,
        feature_data,
        id=f"{feature.uri}-{feature.name}-{pickle.name}{feature.build_pickle_table_rows_breadcrumb(pickle)}",
        marks=marks,
    )


chain_map = compose(chain.from_iterable, map)


def pytest_generate_tests(metafunc: Metafunc):
    config = metafunc.config

    # build marker locators
    marks: Sequence[Mark] = metafunc.definition.own_markers
    mark_names = list(map(attrgetter("name"), marks))
    if "pytest_bdd_scenario" in mark_names:
        scenario_marks = filter(lambda mark: mark.name == "scenarios", marks)
        locators = chain_map(partial(_build_scenario_locators_from_mark, config=config), scenario_marks)
        feature_scenario_feature_source = chain_map(methodcaller("resolve", config), locators)

        metafunc.parametrize(
            "feature, scenario, feature_source",
            starmap(partial(_build_scenario_param, config=config), feature_scenario_feature_source),
        )


def pytest_cmdline_main(config: Config) -> Optional[int]:
    return generation.cmdline_main(config)


def _pytest_collect_file(parent: Collector, file_path=None):
    file_path = Path(file_path)
    config = parent.session.config
    is_enabled_feature_autoload = config.getoption("feature_autoload")
    if is_enabled_feature_autoload is None:
        is_enabled_feature_autoload = not config.getini("disable_feature_autoload")
    if not is_enabled_feature_autoload:
        return

    config = parent.config
    hook = parent.config.hook

    if hook.pytest_bdd_is_collectible(config=config, path=Path(file_path)):
        return FeatureFileCollector.build(parent=parent, file_path=file_path)


if PYTEST7:  # Done intentionally because of API change

    def pytest_collect_file(parent: Collector, file_path):
        return _pytest_collect_file(parent=parent, file_path=file_path)

else:

    def pytest_collect_file(parent: Collector, path):  # type: ignore[misc]
        return _pytest_collect_file(parent=parent, file_path=path)


@pytest.mark.trylast
def pytest_bdd_convert_tag_to_marks(feature, scenario, tag) -> Optional[Collection[Union[Mark, MarkDecorator]]]:
    return [getattr(pytest.mark, tag)]


def pytest_bdd_match_step_definition_to_step(request, feature, scenario, step, previous_step) -> StepHandler.Definition:
    step_registry: StepHandler.Registry = request.getfixturevalue("step_registry")
    step_matcher: StepHandler.Matcher = request.getfixturevalue("step_matcher")

    return step_matcher(request, feature, scenario, step, previous_step, step_registry)


def pytest_bdd_get_mimetype(config: Config, path: Path):
    if str(path).endswith(".gherkin") or str(path).endswith(".feature"):
        return Mimetype.gherkin_plain.value


def pytest_bdd_get_parser(config: Config, mimetype: str):
    return {Mimetype.gherkin_plain.value: GherkinParser}.get(mimetype)


def pytest_bdd_is_collectible(config: Config, path: Path):
    if any(map(partial(contains, {".gherkin", ".feature", ".url", ".desktop", ".webloc"}), path.suffixes)):
        return True



from contextlib import suppress
from functools import partial
from inspect import getmembers
from pathlib import Path

import pytest

from pytest_bdd.compatibility.pytest import PYTEST7, Config, Module
from pytest_bdd.mimetypes import Mimetype
from pytest_bdd.struct_bdd.model import StepPrototype
from pytest_bdd.struct_bdd.parser import StructBDDParser


class StructBDDPlugin:
    extension_to_mimetype = {
        StructBDDParser.KIND.YAML: Mimetype.struct_bdd_yaml,
        StructBDDParser.KIND.HOCON: Mimetype.struct_bdd_hocon,
        StructBDDParser.KIND.JSON5: Mimetype.struct_bdd_json5,
        StructBDDParser.KIND.JSON: Mimetype.struct_bdd_json,
        StructBDDParser.KIND.HJSON: Mimetype.struct_bdd_hjson,
        StructBDDParser.KIND.TOML: Mimetype.struct_bdd_toml,
    }

    def pytest_bdd_get_parser(self, config: Config, mimetype: str):
        with suppress(KeyError):
            return partial(
                StructBDDParser,
                kind={
                    Mimetype.struct_bdd_yaml.value: StructBDDParser.KIND.YAML.value,
                    Mimetype.struct_bdd_hocon.value: StructBDDParser.KIND.HOCON.value,
                    Mimetype.struct_bdd_json5.value: StructBDDParser.KIND.JSON5.value,
                    Mimetype.struct_bdd_json.value: StructBDDParser.KIND.JSON.value,
                    Mimetype.struct_bdd_hjson.value: StructBDDParser.KIND.HJSON.value,
                    Mimetype.struct_bdd_toml.value: StructBDDParser.KIND.TOML.value,
                }[mimetype],
            )

    def pytest_bdd_get_mimetype(self, config: Config, path: Path):
        for extension_suffix, mimetype in self.extension_to_mimetype.items():
            if str(path).endswith(f".bdd.{extension_suffix.value}"):
                return mimetype.value

    def pytest_bdd_is_collectible(self, config: Config, path: Path):
        for extension_suffix, mimetype in self.extension_to_mimetype.items():
            if str(path).endswith(f".bdd.{extension_suffix.value}"):
                return True

    def _pytest_pycollect_makemodule(self):
        outcome = yield
        res = outcome.get_result()
        if isinstance(res, Module):
            for member_name, member in getmembers(res.module):
                if isinstance(member, StepPrototype) and member_name.startswith("test_"):
                    setattr(res.module, member_name, member.as_test(res.module.__file__))

    if PYTEST7:

        @pytest.hookimpl(hookwrapper=True)
        def pytest_pycollect_makemodule(self, parent, module_path):
            yield from self._pytest_pycollect_makemodule()

    else:

        @pytest.hookimpl(hookwrapper=True)
        def pytest_pycollect_makemodule(self, path, parent):
            yield from self._pytest_pycollect_makemodule()
